import sys
from _typeshed import SupportsGetItem, SupportsItemAccess
from typing import IO, Any, AnyStr, Dict, Iterable, Iterator, List, Mapping, Optional, Protocol, Tuple, TypeVar, Union

_T = TypeVar("_T", bound=FieldStorage)

def parse(
    fp: Optional[IO[Any]] = ...,
    environ: SupportsItemAccess[str, str] = ...,
    keep_blank_values: bool = ...,
    strict_parsing: bool = ...,
) -> Dict[str, List[str]]: ...

if sys.version_info < (3, 8):
    def parse_qs(qs: str, keep_blank_values: bool = ..., strict_parsing: bool = ...) -> Dict[str, List[str]]: ...
    def parse_qsl(qs: str, keep_blank_values: bool = ..., strict_parsing: bool = ...) -> Dict[str, List[str]]: ...

if sys.version_info >= (3, 7):
    def parse_multipart(
        fp: IO[Any], pdict: SupportsGetItem[str, bytes], encoding: str = ..., errors: str = ...
    ) -> Dict[str, List[Any]]: ...

else:
    def parse_multipart(fp: IO[Any], pdict: SupportsGetItem[str, bytes]) -> Dict[str, List[bytes]]: ...

class _Environ(Protocol):
    def __getitem__(self, __k: str) -> str: ...
    def keys(self) -> Iterable[str]: ...

def parse_header(line: str) -> Tuple[str, Dict[str, str]]: ...
def test(environ: _Environ = ...) -> None: ...
def print_environ(environ: _Environ = ...) -> None: ...
def print_form(form: Dict[str, Any]) -> None: ...
def print_directory() -> None: ...
def print_environ_usage() -> None: ...

if sys.version_info < (3,):
    def escape(s: AnyStr, quote: bool = ...) -> AnyStr: ...

elif sys.version_info < (3, 8):
    def escape(s: str, quote: Optional[bool] = ...) -> str: ...

class MiniFieldStorage:
    # The first five "Any" attributes here are always None, but mypy doesn't support that
    filename: Any
    list: Any
    type: Any
    file: Optional[IO[bytes]]
    type_options: Dict[Any, Any]
    disposition: Any
    disposition_options: Dict[Any, Any]
    headers: Dict[Any, Any]
    name: Any
    value: Any
    def __init__(self, name: Any, value: Any) -> None: ...
    def __repr__(self) -> str: ...

class FieldStorage(object):
    FieldStorageClass: Optional[type]
    keep_blank_values: int
    strict_parsing: int
    qs_on_post: Optional[str]
    headers: Mapping[str, str]
    fp: IO[bytes]
    encoding: str
    errors: str
    outerboundary: bytes
    bytes_read: int
    limit: Optional[int]
    disposition: str
    disposition_options: Dict[str, str]
    filename: Optional[str]
    file: Optional[IO[bytes]]
    type: str
    type_options: Dict[str, str]
    innerboundary: bytes
    length: int
    done: int
    list: Optional[List[Any]]
    value: Union[None, bytes, List[Any]]

    if sys.version_info >= (3, 6):
        def __init__(
            self,
            fp: Optional[IO[Any]] = ...,
            headers: Optional[Mapping[str, str]] = ...,
            outerboundary: bytes = ...,
            environ: SupportsGetItem[str, str] = ...,
            keep_blank_values: int = ...,
            strict_parsing: int = ...,
            limit: Optional[int] = ...,
            encoding: str = ...,
            errors: str = ...,
            max_num_fields: Optional[int] = ...,
        ) -> None: ...
    elif sys.version_info >= (3, 0):
        def __init__(
            self,
            fp: Optional[IO[Any]] = ...,
            headers: Optional[Mapping[str, str]] = ...,
            outerboundary: bytes = ...,
            environ: SupportsGetItem[str, str] = ...,
            keep_blank_values: int = ...,
            strict_parsing: int = ...,
            limit: Optional[int] = ...,
            encoding: str = ...,
            errors: str = ...,
        ) -> None: ...
    else:
        def __init__(
            self,
            fp: IO[Any] = ...,
            headers: Mapping[str, str] = ...,
            outerboundary: bytes = ...,
            environ: SupportsGetItem[str, str] = ...,
            keep_blank_values: int = ...,
            strict_parsing: int = ...,
        ) -> None: ...
    if sys.version_info >= (3, 0):
        def __enter__(self: _T) -> _T: ...
        def __exit__(self, *args: Any) -> None: ...
    def __repr__(self) -> str: ...
    def __iter__(self) -> Iterator[str]: ...
    def __getitem__(self, key: str) -> Any: ...
    def getvalue(self, key: str, default: Any = ...) -> Any: ...
    def getfirst(self, key: str, default: Any = ...) -> Any: ...
    def getlist(self, key: str) -> List[Any]: ...
    def keys(self) -> List[str]: ...
    if sys.version_info < (3, 0):
        def has_key(self, key: str) -> bool: ...
    def __contains__(self, key: str) -> bool: ...
    def __len__(self) -> int: ...
    if sys.version_info >= (3, 0):
        def __bool__(self) -> bool: ...
    else:
        def __nonzero__(self) -> bool: ...
    if sys.version_info >= (3, 0):
        # In Python 3 it returns bytes or str IO depending on an internal flag
        def make_file(self) -> IO[Any]: ...
    else:
        # In Python 2 it always returns bytes and ignores the "binary" flag
        def make_file(self, binary: Any = ...) -> IO[bytes]: ...

if sys.version_info < (3, 0):
    from UserDict import UserDict
    class FormContentDict(UserDict[str, List[str]]):
        query_string: str
        def __init__(self, environ: Mapping[str, str] = ..., keep_blank_values: int = ..., strict_parsing: int = ...) -> None: ...
    class SvFormContentDict(FormContentDict):
        def getlist(self, key: Any) -> Any: ...
    class InterpFormContentDict(SvFormContentDict): ...
    class FormContent(FormContentDict):
        # TODO this should have
        # def values(self, key: Any) -> Any: ...
        # but this is incompatible with the supertype, and adding '# type: ignore' triggers
        # a parse error in pytype (https://github.com/google/pytype/issues/53)
        def indexed_value(self, key: Any, location: int) -> Any: ...
        def value(self, key: Any) -> Any: ...
        def length(self, key: Any) -> int: ...
        def stripped(self, key: Any) -> Any: ...
        def pars(self) -> Dict[Any, Any]: ...



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import fnmatch
import json
import logging
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Set

from .. import log
from ..analysis_directory import AnalysisDirectory
from ..configuration import Configuration
from ..error import Error
from .command import TEXT, ClientException, Command, CommandArguments, Result


LOG: logging.Logger = logging.getLogger(__name__)


class Reporting(Command):
    NAME = "reporting"

    def __init__(
        self,
        command_arguments: CommandArguments,
        original_directory: str,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
    ) -> None:
        super().__init__(
            command_arguments, original_directory, configuration, analysis_directory
        )

    def _print(self, errors: Sequence[Error]) -> None:
        if errors:
            length = len(errors)
            LOG.error("Found %d type error%s!", length, "s" if length > 1 else "")
        else:
            LOG.log(log.SUCCESS, "No type errors found")

        if self._output == TEXT:
            log.stdout.write("\n".join([repr(error) for error in errors]))
        else:
            log.stdout.write(json.dumps([error.__dict__ for error in errors]))

    def _get_directories_to_analyze(self) -> Set[str]:
        return self._analysis_directory.get_filter_roots()

    @staticmethod
    def _load_errors_from_json(json_output: str) -> List[Dict[str, Any]]:
        try:
            json_dictionary = json.loads(json_output)
        except (json.JSONDecodeError):
            raise ClientException(f"Invalid JSON output: `{json_output}`.")

        error_list = []
        if isinstance(json_dictionary, dict) and "errors" in json_dictionary:
            error_list = json_dictionary["errors"]
        else:
            # TODO(T62259082): Identify why the client receives such JSON output.
            LOG.error(
                "Received invalid JSON output for the last `pyre` command"
                " (known bug: GitHub issue #238 / T62259082)."
                " You may want to rerun your command."
            )
            LOG.debug("Invalid JSON output: %s", json_output)
        return error_list

    def _get_errors(
        self, result: Result, bypass_filtering: bool = False
    ) -> Sequence[Error]:
        result.check()
        errors: List[Error] = []
        results: List[Dict[str, Any]] = self._load_errors_from_json(result.output)

        for error in results:
            path = os.path.realpath(error["path"])
            analysis_root = os.path.realpath(self._analysis_directory.get_root())
            if not Path(analysis_root) in Path(path).parents:
                path = os.path.realpath(os.path.join(analysis_root, error["path"]))

            # Relativize path to user's cwd.
            relative_path = self._relative_path(path)
            error["path"] = relative_path
            ignore_error = False
            external_to_global_root = True
            if path.startswith(self._project_root):
                external_to_global_root = False
            if not os.path.exists(path):
                # Nonexistent paths can be created when search path stubs are renamed.
                external_to_global_root = True
            for absolute_ignore_path in self._ignore_all_errors_paths:
                if fnmatch.fnmatch(path, (absolute_ignore_path + "*")):
                    ignore_error = True
                    break
            errors.append(Error(error, ignore_error, external_to_global_root))

        if bypass_filtering:
            return errors
        else:
            filtered_errors = [
                error
                for error in errors
                if (
                    not error.is_ignored()
                    and (not (error.is_external_to_global_root()))
                )
            ]
            sorted_errors = sorted(
                filtered_errors,
                key=lambda error: (error.path, error.line, error.column),
            )

            return sorted_errors



from __future__ import annotations

from collections import defaultdict
from enum import Enum
from typing import Dict, List, Optional, Set, Tuple, Union

from sqlalchemy.orm import Session, aliased
from sqlalchemy.orm.query import Query
from sqlalchemy.sql.expression import or_

from .models import (
    Issue,
    IssueInstance,
    IssueInstanceSharedTextAssoc,
    SharedText,
    SharedTextKind,
)


FilenameText = aliased(SharedText)
CallableText = aliased(SharedText)
CallerText = aliased(SharedText)
CalleeText = aliased(SharedText)
MessageText = aliased(SharedText)


class Filter(Enum):
    codes = "codes"
    callables = "callables"
    file_names = "file_names"
    trace_length_to_sources = "trace_length_to_sources"
    trace_length_to_sinks = "trace_length_to_sinks"
    any_features = "any_features"
    all_features = "all_features"
    exclude_features = "exclude_features"


class IssueQueryBuilder:
    def __init__(self, current_run_id: int):
        self._session = None
        self.current_run_id = current_run_id
        self.issue_filters: Dict[
            Filter, Set[Tuple[Union[int, str, Tuple[int, int, ...]], ...]]
        ] = defaultdict(set)
        self.breadcrumb_filters: Dict[Filter, List[str]] = defaultdict(list)

    @property
    def session(self):
        if self._session is None:
            raise Exception("No current session found for query!")
        return self._session

    def with_session(self, session):
        self._session = session
        return self

    def get(self) -> List:
        query = self._get_session_query(self._session)
        for filter_type, filter_conditions in self.issue_filters.items():
            if filter_type == Filter.codes:
                column = Issue.code
            elif filter_type == Filter.callables:
                column = CallableText.contents
            elif filter_type == Filter.file_names:
                column = FilenameText.contents
            elif filter_type == Filter.trace_length_to_sources:
                column = IssueInstance.min_trace_length_to_sources
            elif filter_type == Filter.trace_length_to_sinks:
                column = IssueInstance.min_trace_length_to_sinks

            for filter_condition in filter_conditions:
                if (
                    filter_type == Filter.trace_length_to_sources
                    or filter_type == Filter.trace_length_to_sinks
                ):
                    if filter_condition[0]:
                        query = query.filter(column >= filter_condition[0])
                    if filter_condition[1]:
                        query = query.filter(column <= filter_condition[1])
                else:
                    query = query.filter(
                        or_(*[column.like(item) for item in filter_condition])
                    )
        issues = list(
            query.join(Issue, IssueInstance.issue_id == Issue.id).join(
                MessageText, MessageText.id == IssueInstance.message_id
            )
        )

        any_feature_set = set()
        all_feature_set = set()
        exclude_feature_set = set()
        for filter_type, filter_condition in self.breadcrumb_filters.items():
            if filter_type == Filter.any_features:
                any_feature_set |= set(filter_condition)
            elif filter_type == Filter.all_features:
                all_feature_set |= set(filter_condition)
            elif filter_type == Filter.exclude_features:
                exclude_feature_set |= set(filter_condition)
            else:
                raise Exception(f"Invalid filter type provided: {filter_type}")

            features_list = [
                self._get_leaves_issue_instance(
                    self._session, int(issue.id), SharedTextKind.FEATURE
                )
                for issue in issues
            ]
            for issue, features in zip(issues, features_list):
                if any_feature_set and not (features & any_feature_set):
                    issues.remove(issue)
                elif all_feature_set and not (
                    features & all_feature_set == all_feature_set
                ):
                    issues.remove(issue)
                elif exclude_feature_set and features & exclude_feature_set:
                    issues.remove(issue)
        return issues

    def where_codes_is_any_of(self, codes: List[int]) -> IssueQueryBuilder:
        self.issue_filters[Filter.codes].add(tuple(codes))
        return self

    def where_callables_is_any_of(self, callables: List[str]) -> IssueQueryBuilder:
        self.issue_filters[Filter.callables].add(tuple(callables))
        return self

    def where_file_names_is_any_of(self, file_names: List[str]) -> IssueQueryBuilder:
        self.issue_filters[Filter.file_names].add(tuple(file_names))
        return self

    def where_trace_length_to_sinks(
        self, minimum: Optional[int] = None, maximum: Optional[int] = None
    ) -> IssueQueryBuilder:
        self.issue_filters[Filter.trace_length_to_sinks].add((minimum, maximum))
        return self

    def where_trace_length_to_sources(
        self, minimum: Optional[int] = None, maximum: Optional[int] = None
    ) -> IssueQueryBuilder:
        self.issue_filters[Filter.trace_length_to_sources].add((minimum, maximum))
        return self

    def where_any_features(self, features: List[str]) -> IssueQueryBuilder:
        self.breadcrumb_filters[Filter.any_features] += features
        return self

    def where_all_features(self, features: List[str]) -> IssueQueryBuilder:
        self.breadcrumb_filters[Filter.all_features] += features
        return self

    def where_exclude_features(self, features: List[str]) -> IssueQueryBuilder:
        self.breadcrumb_filters[Filter.exclude_features] += features
        return self

    def sources(self, issues) -> List[Set[str]]:
        return [
            self._get_leaves_issue_instance(
                self._session, int(issue.id), SharedTextKind.SINK
            )
            for issue in issues
        ]

    def sinks(self, issues) -> List[Set[str]]:
        return [
            self._get_leaves_issue_instance(
                self._session, int(issue.id), SharedTextKind.SOURCE
            )
            for issue in issues
        ]

    def features(self, issues) -> List[Set[str]]:
        return [
            self._get_leaves_issue_instance(
                self._session, int(issue.id), SharedTextKind.FEATURE
            )
            for issue in issues
        ]

    def _get_session_query(self, session: Session) -> Query:
        return (
            session.query(
                IssueInstance.id,
                FilenameText.contents.label("filename"),
                IssueInstance.location,
                Issue.code,
                CallableText.contents.label("callable"),
                MessageText.contents.label("message"),
                IssueInstance.min_trace_length_to_sources,
                IssueInstance.min_trace_length_to_sinks,
            )
            .filter(IssueInstance.run_id == self.current_run_id)
            .join(FilenameText, FilenameText.id == IssueInstance.filename_id)
            .join(CallableText, CallableText.id == IssueInstance.callable_id)
        )

    def _get_leaves_issue_instance(
        self, session: Session, issue_instance_id: int, kind: SharedTextKind
    ) -> Set[str]:
        message_ids = [
            int(id)
            for id, in session.query(SharedText.id)
            .distinct(SharedText.id)
            .join(
                IssueInstanceSharedTextAssoc,
                SharedText.id == IssueInstanceSharedTextAssoc.shared_text_id,
            )
            .filter(IssueInstanceSharedTextAssoc.issue_instance_id == issue_instance_id)
            .filter(SharedText.kind == kind)
        ]
        return self._leaf_dict_lookups(message_ids, kind, session)

    def _leaf_dict_lookups(
        self, message_ids: List[int], kind: SharedTextKind, session: Session
    ) -> Set[str]:
        if kind == SharedTextKind.SOURCE:
            leaf_dict = {
                int(id): contents
                for id, contents in session.query(
                    SharedText.id, SharedText.contents
                ).filter(SharedText.kind == kind)
            }
        elif kind == SharedTextKind.SINK:
            leaf_dict = {
                int(id): contents
                for id, contents in session.query(
                    SharedText.id, SharedText.contents
                ).filter(SharedText.kind == kind)
            }
        else:
            leaf_dict = {
                int(id): contents
                for id, contents in session.query(
                    SharedText.id, SharedText.contents
                ).filter(SharedText.kind == kind)
            }
        return {leaf_dict[id] for id in message_ids if id in leaf_dict}



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
from typing import Optional

from ..analysis_directory import AnalysisDirectory, resolve_analysis_directory
from ..configuration import Configuration
from .command import Command, CommandArguments, ExitCode, IncrementalStyle
from .incremental import Incremental
from .start import Start  # noqa
from .stop import Stop


class Restart(Command):
    NAME = "restart"

    def __init__(
        self,
        command_arguments: CommandArguments,
        original_directory: str,
        *,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
        terminal: bool,
        store_type_check_resolution: bool,
        use_watchman: bool,
        incremental_style: IncrementalStyle,
    ) -> None:
        super(Restart, self).__init__(
            command_arguments, original_directory, configuration, analysis_directory
        )
        self._terminal: bool = terminal
        self._store_type_check_resolution: bool = store_type_check_resolution
        self._use_watchman: bool = use_watchman
        self._incremental_style: IncrementalStyle = incremental_style

    @staticmethod
    def from_arguments(
        arguments: argparse.Namespace,
        original_directory: str,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
    ) -> "Restart":
        return Restart(
            CommandArguments.from_arguments(arguments),
            original_directory,
            configuration=configuration,
            analysis_directory=analysis_directory,
            terminal=arguments.terminal,
            store_type_check_resolution=arguments.store_type_check_resolution,
            use_watchman=not arguments.no_watchman,
            incremental_style=arguments.incremental_style,
        )

    @classmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> None:
        restart = parser.add_parser(
            cls.NAME, epilog="Restarts a server. Equivalent to `pyre stop && pyre`."
        )
        restart.set_defaults(command=cls.from_arguments)
        restart.add_argument(
            "--terminal", action="store_true", help="Run the server in the terminal."
        )
        restart.add_argument(
            "--store-type-check-resolution",
            action="store_true",
            help="Store extra information for `types` queries.",
        )
        restart.add_argument(
            "--no-watchman",
            action="store_true",
            help="Do not spawn a watchman client in the background.",
        )
        restart.add_argument(
            "--incremental-style",
            type=IncrementalStyle,
            choices=list(IncrementalStyle),
            default=IncrementalStyle.FINE_GRAINED,
            help="How to approach doing incremental checks.",
        )

    def generate_analysis_directory(self) -> AnalysisDirectory:
        return resolve_analysis_directory(
            self._source_directories,
            self._targets,
            self._configuration,
            self._original_directory,
            self._project_root,
            filter_directory=self._filter_directory,
            use_buck_builder=self._use_buck_builder,
            debug=self._debug,
            buck_mode=self._buck_mode,
            relative_local_root=self.relative_local_root,
        )

    def _run(self) -> None:
        exit_code = (
            Stop(
                self._command_arguments,
                self._original_directory,
                configuration=self._configuration,
                analysis_directory=self._analysis_directory,
                from_restart=True,
            )
            .run()
            .exit_code()
        )
        if exit_code != ExitCode.SUCCESS:
            self._exit_code = ExitCode.FAILURE
            return
        exit_code = (
            Incremental(
                self._command_arguments,
                self._original_directory,
                configuration=self._configuration,
                analysis_directory=self._analysis_directory,
                # Force the incremental run to be blocking.
                nonblocking=False,
                incremental_style=self._incremental_style,
                no_start_server=False,
                no_watchman=not self._use_watchman,
            )
            .run()
            .exit_code()
        )
        self._exit_code = exit_code



# pyre-unsafe

import datetime
from typing import Any, Dict, Optional

class Widget:
    needs_multipart_form: bool = ...
    is_localized: bool = ...
    is_required: bool = ...
    supports_microseconds: bool = ...
    def __init__(self, attrs: Optional[Dict[str, Any]] = ...) -> None: ...

class Input(Widget):
    input_type: str = ...
    template_name: str = ...

CheckboxInput = Input
EmailInput = Input
TextInput = Input
NumberInput = Input
URLInput = Input
HiddenInput = Input



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import argparse
import enum
import json
import logging
import os
import resource
import signal
import subprocess
import threading
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import IO, Iterable, List, Optional

from typing_extensions import Final

from .. import json_rpc, log, recently_used_configurations, terminal
from ..analysis_directory import AnalysisDirectory, resolve_analysis_directory
from ..configuration import Configuration
from ..exceptions import EnvironmentException
from ..filesystem import readable_directory, remove_if_exists, translate_path
from ..find_directories import (
    LOCAL_CONFIGURATION_FILE,
    find_local_root,
    find_project_root,
)
from ..log import StreamLogger
from ..process import Process
from ..resources import LOG_DIRECTORY, find_log_directory
from ..socket_connection import SocketConnection, SocketException


TEXT: str = "text"
JSON: str = "json"


LOG: logging.Logger = logging.getLogger(__name__)


class ClientException(Exception):
    pass


class State(enum.IntEnum):
    DEAD = 0
    RUNNING = 1


class ExitCode(enum.IntEnum):
    SUCCESS = 0
    FOUND_ERRORS = 1
    FAILURE = 2
    BUCK_ERROR = 3
    # If the process exited due to a signal, this will be the negative signal number.
    SIGSEGV = -signal.SIGSEGV


class IncrementalStyle(enum.Enum):
    SHALLOW = "shallow"
    FINE_GRAINED = "fine_grained"

    def __str__(self) -> str:
        return self.value


class ProfileOutput(enum.Enum):
    TRACE_EVENT: str = "trace_event"
    COLD_START_PHASES: str = "cold_start_phases"
    INCREMENTAL_UPDATES: str = "incremental_updates"
    INDIVIDUAL_TABLE_SIZES: str = "individual_table_sizes"
    TOTAL_SHARED_MEMORY_SIZE_OVER_TIME: str = "total_shared_memory_size_over_time"
    TOTAL_SHARED_MEMORY_SIZE_OVER_TIME_GRAPH: str = "total_shared_memory_size_over_time_graph"  # noqa B950

    def __str__(self) -> str:
        return self.value


class Result:
    def __init__(self, code: int, output: str, error: Optional[str] = None) -> None:
        self.code: int = code
        self.output: str = output
        self.error: Optional[str] = error

    def check(self) -> None:
        if self.code != ExitCode.SUCCESS:
            description = ":\n{}".format(self.output) if self.output else ""
            if self.code == ExitCode.SIGSEGV:
                description += (
                    "\nThis is a Pyre bug. Please re-run Pyre with --debug "
                    "and provide the output to the developers."
                )
            raise ClientException(
                "Client exited with error code {}{}".format(self.code, description)
            )


def typeshed_search_path(typeshed_root: str) -> List[str]:
    search_path = []
    typeshed_subdirectories = ["stdlib", "third_party"]
    for typeshed_subdirectory_name in typeshed_subdirectories:
        typeshed_subdirectory = os.path.join(typeshed_root, typeshed_subdirectory_name)
        if (
            not os.path.isdir(typeshed_subdirectory)
            or typeshed_subdirectory_name == "tests"
            or typeshed_subdirectory_name[0] == "."
        ):
            continue

        # Always prefer newer version over older version
        # pyre-fixme[6]: Expected `Iterable[Variable[_LT (bound to
        #  _SupportsLessThan)]]` for 1st param but got `List[str]`.
        version_names = sorted(os.listdir(typeshed_subdirectory), reverse=True)
        for version_name in version_names:
            # Anything under 2/ or 2.x is unusable for Pyre
            if version_name.startswith("2") and version_name != "2and3":
                continue
            search_path.append(os.path.join(typeshed_subdirectory, version_name))
    return search_path


def _convert_json_response_to_result(response: json_rpc.Response) -> Result:
    error = response.error
    error_message = None
    if error:
        error_code = ExitCode.FAILURE
        error_message = json.dumps(error)
    else:
        error_code = ExitCode.SUCCESS
    return Result(
        output=json.dumps(response.result), code=error_code, error=error_message
    )


def executable_file(file_path: str) -> str:
    if not os.path.isfile(file_path):
        raise EnvironmentException("%s is not a valid file" % file_path)
    if not os.access(file_path, os.X_OK):
        raise EnvironmentException("%s is not an executable file" % file_path)
    return file_path


@dataclass(frozen=True)
class CommandArguments:
    local_configuration: Optional[str]
    version: bool
    debug: bool
    sequential: bool
    strict: bool
    additional_checks: List[str]
    show_error_traces: bool
    output: str
    enable_profiling: bool
    enable_memory_profiling: bool
    noninteractive: bool
    hide_parse_errors: bool
    logging_sections: Optional[str]
    log_identifier: str
    logger: Optional[str]
    formatter: List[str]
    targets: List[str]
    use_buck_builder: Optional[bool]
    source_directories: List[str]
    filter_directory: Optional[str]
    buck_mode: Optional[str]
    no_saved_state: bool
    search_path: List[str]
    binary: str
    buck_builder_binary: Optional[str]
    exclude: List[str]
    typeshed: str
    save_initial_state_to: Optional[str]
    load_initial_state_from: Optional[str]
    changed_files_path: Optional[str]
    saved_state_project: Optional[str]
    dot_pyre_directory: Optional[Path]
    features: Optional[str]

    @staticmethod
    def from_arguments(arguments: argparse.Namespace) -> "CommandArguments":
        return CommandArguments(
            local_configuration=arguments.local_configuration,
            version=arguments.version,
            debug=arguments.debug,
            sequential=arguments.sequential,
            strict=arguments.strict,
            additional_checks=arguments.additional_check,
            show_error_traces=arguments.show_error_traces,
            output=arguments.output,
            enable_profiling=arguments.enable_profiling,
            enable_memory_profiling=arguments.enable_memory_profiling,
            noninteractive=arguments.noninteractive,
            hide_parse_errors=arguments.hide_parse_errors,
            logging_sections=arguments.logging_sections,
            log_identifier=arguments.log_identifier,
            logger=arguments.logger,
            formatter=arguments.formatter,
            targets=arguments.targets,
            use_buck_builder=arguments.use_buck_builder,
            source_directories=arguments.source_directories,
            filter_directory=arguments.filter_directory,
            buck_mode=arguments.buck_mode,
            no_saved_state=arguments.no_saved_state,
            search_path=arguments.search_path,
            binary=arguments.binary,
            buck_builder_binary=arguments.buck_builder_binary,
            exclude=arguments.exclude,
            typeshed=arguments.typeshed,
            save_initial_state_to=arguments.save_initial_state_to,
            load_initial_state_from=arguments.load_initial_state_from,
            changed_files_path=arguments.changed_files_path,
            saved_state_project=arguments.saved_state_project,
            dot_pyre_directory=arguments.dot_pyre_directory,
            features=arguments.features,
        )


class CommandParser(ABC):
    NAME = ""
    HIDDEN = False
    _exit_code: ExitCode = ExitCode.SUCCESS

    def __init__(
        self, command_arguments: CommandArguments, original_directory: str
    ) -> None:
        self._command_arguments = command_arguments

        self._version: bool = self._command_arguments.version
        self._debug: bool = self._command_arguments.debug
        self._sequential: bool = self._command_arguments.sequential
        self._strict: bool = self._command_arguments.strict
        self._additional_checks: List[str] = self._command_arguments.additional_checks
        self._show_error_traces: bool = self._command_arguments.show_error_traces
        self._output: str = self._command_arguments.output
        self._enable_profiling: bool = self._command_arguments.enable_profiling
        self._enable_memory_profiling: bool = (
            self._command_arguments.enable_memory_profiling
        )
        self._noninteractive: bool = self._command_arguments.noninteractive
        self._hide_parse_errors: bool = self._command_arguments.hide_parse_errors
        self._log_identifier: str = self._command_arguments.log_identifier
        self._formatter: List[str] = self._command_arguments.formatter

        self._targets: List[str] = self._command_arguments.targets
        self._use_buck_builder: Optional[
            bool
        ] = self._command_arguments.use_buck_builder

        self._source_directories: List[str] = self._command_arguments.source_directories
        self._filter_directory: Optional[str] = self._command_arguments.filter_directory
        self._buck_mode: Optional[str] = self._command_arguments.buck_mode
        self._no_saved_state: bool = self._command_arguments.no_saved_state

        self._search_path: List[str] = self._command_arguments.search_path
        self._binary: str = self._command_arguments.binary
        self._buck_builder_binary: Final[
            Optional[str]
        ] = self._command_arguments.buck_builder_binary
        self._exclude: List[str] = self._command_arguments.exclude
        self._typeshed: str = self._command_arguments.typeshed
        self._save_initial_state_to: Final[
            Optional[str]
        ] = self._command_arguments.save_initial_state_to
        self._load_initial_state_from: Final[
            Optional[str]
        ] = self._command_arguments.load_initial_state_from
        self._changed_files_path: Final[
            Optional[str]
        ] = self._command_arguments.changed_files_path
        self._saved_state_project: Final[
            Optional[str]
        ] = self._command_arguments.saved_state_project

        # Derived arguments
        self._capable_terminal: bool = terminal.is_capable()
        self._original_directory: str = original_directory
        self._project_root: str = find_project_root(self._original_directory)

        local_root = self._command_arguments.local_configuration
        if local_root and local_root.endswith(LOCAL_CONFIGURATION_FILE):
            local_root = local_root[: -len(LOCAL_CONFIGURATION_FILE)]
        self._local_root: Final[Optional[str]] = find_local_root(
            self._original_directory, local_root
        )
        self._dot_pyre_directory: Path = (
            self._command_arguments.dot_pyre_directory
            or Path(self._project_root, LOG_DIRECTORY)
        )
        self._log_directory: str = find_log_directory(
            self._project_root, self._local_root, str(self._dot_pyre_directory)
        )
        Path(self._log_directory).mkdir(parents=True, exist_ok=True)

        self._logging_sections: Optional[str] = self._command_arguments.logging_sections
        self._noninteractive: bool = self._command_arguments.noninteractive
        if self._debug or not self._capable_terminal:
            self._noninteractive = True

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "-l", "--local-configuration", type=str, help="Use a local configuration"
        )

        parser.add_argument(
            "--version",
            action="store_true",
            help="Print the client and binary versions of Pyre.",
        )

        parser.add_argument("--debug", action="store_true", help=argparse.SUPPRESS)
        parser.add_argument("--sequential", action="store_true", help=argparse.SUPPRESS)
        parser.add_argument("--strict", action="store_true", help=argparse.SUPPRESS)
        parser.add_argument(
            "--additional-check", action="append", help=argparse.SUPPRESS
        )

        parser.add_argument(
            "--show-error-traces",
            action="store_true",
            help="Display errors trace information",
        )

        # Logging.
        parser.add_argument(
            "--output", choices=[TEXT, JSON], default=TEXT, help="How to format output"
        )
        parser.add_argument(
            "--enable-profiling", action="store_true", help=argparse.SUPPRESS
        )
        parser.add_argument(
            "--enable-memory-profiling", action="store_true", help=argparse.SUPPRESS
        )
        parser.add_argument(
            "-n",
            "--noninteractive",
            action="store_true",
            help="Disable interactive logging",
        )
        parser.add_argument(
            "--hide-parse-errors",
            action="store_true",
            help="Hide detailed information about parse errors",
        )
        parser.add_argument(
            "--logging-sections", help=argparse.SUPPRESS  # Enable sectional logging.
        )
        parser.add_argument(
            "--log-identifier",
            default="",
            help=argparse.SUPPRESS,  # Add given identifier to logged samples.
        )

        # Directory where Pyre places its log files and artifacts.
        parser.add_argument("--dot-pyre-directory", type=Path, help=argparse.SUPPRESS)
        parser.add_argument(
            "--logger", help=argparse.SUPPRESS  # Specify custom logging binary.
        )
        parser.add_argument("--formatter", help=argparse.SUPPRESS)

        # Link tree determination.
        buck_arguments = parser.add_argument_group("buck")
        buck_arguments.add_argument(
            "--target",
            action="append",
            dest="targets",
            default=[],
            help="The buck target to check",
        )

        use_buck_builder_group = buck_arguments.add_mutually_exclusive_group()
        use_buck_builder_group.add_argument(
            "--use-buck-builder",
            action="store_const",
            const=True,
            help="Use Pyre's experimental builder for Buck projects.",
        )
        use_buck_builder_group.add_argument(
            "--use-legacy-buck-builder",
            dest="use_buck_builder",
            action="store_const",
            const=False,
            help="Use the legacy builder for Buck projects.",
        )

        buck_arguments.add_argument(
            "--buck-mode", type=str, help="Mode to pass to `buck query`"
        )

        source_directories = parser.add_argument_group("source-directories")
        source_directories.add_argument(
            "--source-directory",
            action="append",
            dest="source_directories",
            default=[],
            help="The source directory to check",
            type=os.path.abspath,
        )
        source_directories.add_argument(
            "--filter-directory", help=argparse.SUPPRESS  # override filter directory
        )
        parser.add_argument(
            "--no-saved-state",
            action="store_true",
            help="Don't attempt to load Pyre from a saved state.",
        )

        # Handling of search path
        parser.add_argument(
            "--search-path",
            action="append",
            default=[],
            type=readable_directory,
            help="Add an additional directory of modules and stubs to include"
            " in the type environment",
        )

        parser.add_argument(
            "--binary",
            default=None,
            type=executable_file,
            help="Location of the pyre binary",
        )

        parser.add_argument(
            "--buck-builder-binary",
            default=None,
            help="Location of the buck builder binary",
        )

        parser.add_argument(
            "--exclude",
            action="append",
            default=[],
            help="Exclude files and directories matching this regexp from parsing",
        )

        # Typeshed stubs location
        parser.add_argument(
            "--typeshed",
            default=None,
            type=readable_directory,
            help="Location of the typeshed stubs",
        )
        parser.add_argument(
            "--save-initial-state-to",
            default=None,
            help="Path to serialize pyre's initial state to.",
        )
        parser.add_argument(
            "--load-initial-state-from", default=None, type=str, help=argparse.SUPPRESS
        )
        parser.add_argument(
            "--changed-files-path", default=None, type=str, help=argparse.SUPPRESS
        )
        parser.add_argument(
            "--saved-state-project", default=None, type=str, help=argparse.SUPPRESS
        )
        parser.add_argument(
            "--features", default=None, type=str, help=argparse.SUPPRESS
        )
        # Temporary flag to help migrate to json sockets for incremental and query
        # commands.
        parser.add_argument(
            "--use-json-sockets",
            action="store_true",
            default=False,
            help=argparse.SUPPRESS,
        )

    @classmethod
    @abstractmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> None:
        pass

    @abstractmethod
    def _run(self) -> None:
        """ Abstract method expected to be overridden by subclasses. """
        pass

    def run(self) -> "CommandParser":
        self._run()
        return self

    def cleanup(self) -> None:
        pass

    def exit_code(self) -> ExitCode:
        return self._exit_code

    def result(self) -> Optional[Result]:
        return None

    @property
    def configuration(self) -> Optional[Configuration]:
        return None

    @property
    def project_root(self) -> Optional[str]:
        return self._project_root

    @property
    def local_root(self) -> Optional[str]:
        return self._local_root

    @property
    def log_directory(self) -> str:
        return self._log_directory

    @property
    def noninteractive(self) -> bool:
        return self._noninteractive

    @property
    def relative_local_root(self) -> Optional[str]:
        if not self.local_root:
            return None
        return str(Path(self.log_directory).relative_to(self._dot_pyre_directory))


class Command(CommandParser, ABC):
    _buffer: List[str] = []

    def __init__(
        self,
        command_arguments: CommandArguments,
        original_directory: str,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
    ) -> None:
        super(Command, self).__init__(command_arguments, original_directory)
        logger = self._command_arguments.logger
        if logger:
            logger = translate_path(self._original_directory, logger)
        self._configuration: Configuration = (
            configuration or self.generate_configuration(logger)
        )
        self._strict: bool = (
            self._command_arguments.strict or self._configuration.strict
        )
        self._logger: Final[Optional[str]] = logger or (self._configuration.logger)
        self._ignore_all_errors_paths: Iterable[str] = (
            self._configuration.ignore_all_errors
        )
        self._number_of_workers: int = self._configuration.number_of_workers
        self._version_hash: str = self._configuration.version_hash
        self._formatter: Final[Optional[str]] = self._configuration.formatter
        self._taint_models_path: List[str] = [
            translate_path(self._original_directory, path)
            for path in self._configuration.taint_models_path
        ]

        self._analysis_directory: AnalysisDirectory = (
            analysis_directory or self.generate_analysis_directory()
        )
        self._features: Final[Optional[str]] = self._command_arguments.features

    @classmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> None:
        pass

    def generate_configuration(self, logger: Optional[str]) -> Configuration:
        return Configuration(
            project_root=self._project_root,
            local_root=self._local_root,
            search_path=self._search_path,
            binary=self._binary,
            typeshed=self._typeshed,
            buck_builder_binary=self._buck_builder_binary,
            excludes=self._exclude,
            logger=logger,
            formatter=self._formatter,
            log_directory=self._log_directory,
        )

    def generate_analysis_directory(self) -> AnalysisDirectory:
        configuration = self._configuration
        if not configuration:
            return AnalysisDirectory(".")
        else:
            return resolve_analysis_directory(
                self._source_directories,
                self._targets,
                configuration,
                self._original_directory,
                self._project_root,
                filter_directory=self._filter_directory,
                use_buck_builder=self._use_buck_builder,
                debug=self._debug,
                buck_mode=self._buck_mode,
                relative_local_root=self.relative_local_root,
            )

    def run(self) -> "Command":
        configuration = self._configuration
        if configuration and configuration.disabled:
            LOG.log(log.SUCCESS, "Pyre will not run due to being explicitly disabled")
        else:
            relative_local_root = self.relative_local_root
            if relative_local_root:
                recently_used_configurations.Cache(self._dot_pyre_directory).put(
                    relative_local_root
                )
            self._run()
        return self

    def _run(self) -> None:
        pass

    def cleanup(self) -> None:
        self._analysis_directory.cleanup()

    def _flags(self) -> List[str]:
        flags = []
        if self._debug:
            flags.extend(["-debug"])
        if self._sequential:
            flags.extend(["-sequential"])
        if self._strict:
            flags.extend(["-strict"])
        if self._additional_checks:
            flags.append("-additional-checks")
            flags.append(",".join(self._additional_checks))
        if self._show_error_traces:
            flags.append("-show-error-traces")
        if not self._hide_parse_errors:
            self._enable_logging_section("parser")
        logging_sections = self._logging_sections
        if not self._capable_terminal:
            # Disable progress reporting for non-capable terminals.
            # This helps in reducing clutter.
            if logging_sections:
                logging_sections = logging_sections + ",-progress"
            else:
                logging_sections = "-progress"
        if logging_sections:
            flags.extend(["-logging-sections", logging_sections])
        if self._enable_profiling:
            flags.extend(["-profiling-output", self.profiling_log_path()])
        if self._enable_memory_profiling:
            flags.extend(["-memory-profiling-output", self.profiling_log_path()])
        if self._enable_profiling or self._enable_memory_profiling:
            # Clear the profiling log first since in pyre binary it's append-only
            remove_if_exists(self.profiling_log_path())
        if self._project_root:
            flags.extend(["-project-root", self._project_root])
        if self._log_identifier:
            flags.extend(["-log-identifier", self._log_identifier])
        if self._logger:
            flags.extend(["-logger", self._logger])
        if self._log_directory:
            flags.extend(["-log-directory", self._log_directory])
        return flags

    # temporarily always return empty list to unblock client release
    def _feature_flags(self) -> List[str]:
        features = self._features
        if features:
            lsp_features = ["click_to_fix", "hover", "go_to_definition"]
            filtered = {
                key: value
                for key, value in json.loads(features).items()
                if key in lsp_features
            }
            return ["-features", json.dumps(filtered)]
        return []

    def _read_stdout(self, stdout: Iterable[str]) -> None:
        self._buffer = []
        for line in stdout:
            self._buffer.append(line)

    def _call_client(
        self,
        command: str,
        capture_output: bool = True,
        stdout: Optional[IO[str]] = None,
    ) -> Result:
        if not os.path.isdir(self._analysis_directory.get_root()):
            raise EnvironmentException(
                "`{}` is not a link tree.".format(self._analysis_directory.get_root())
            )

        client_command = [self._configuration.binary, command]
        client_command.extend(self._flags())
        client_command.append(self._analysis_directory.get_root())

        def limit_memory_usage() -> None:
            try:
                limit = 30 * 1024 * 1024 * 1024  # 30 GB
                resource.setrlimit(resource.RLIMIT_DATA, (limit, limit))
            except OSError:
                # Run the process with unlimited memory if the underlying syscall fails.
                pass

        if capture_output:
            assert stdout is None, "capture_output and stdout are mutually exclusive"
            stdout = subprocess.PIPE

        LOG.debug("Running `%s`", " ".join(client_command))
        with subprocess.Popen(
            client_command,
            stdout=stdout,
            stderr=subprocess.PIPE,
            preexec_fn=limit_memory_usage,
            universal_newlines=True,
        ) as process:

            # Read stdout output
            if capture_output:
                stdout_reader = threading.Thread(
                    target=self._read_stdout, args=(process.stdout,)
                )
                stdout_reader.daemon = True
                stdout_reader.start()

            # Read the error output and print it.
            # pyre-fixme[6]: Expected `Iterable[str]` for 1st param but got
            #  `Optional[IO[typing.Any]]`.
            with StreamLogger(process.stderr) as stream_logger:
                with Process.register_non_unique_process(
                    process.pid, self.NAME, self.log_directory
                ):
                    # Wait for the process to finish and clean up.
                    process.wait()
                # In the exceptional case, make sure that we don't stop early
                if process.returncode != 0:
                    stream_logger.join()

            if capture_output:
                stdout_reader.join()

            output = ""
            if capture_output:
                output = "".join(self._buffer)
            if process.returncode != 0 and capture_output:
                output = "".join(self._buffer)

            return Result(code=process.returncode, output=output)

    def _relative_path(self, path: str) -> str:
        return os.path.relpath(path, self._original_directory)

    def _state(self) -> State:
        pid_path = Path(self._log_directory, "server/server.pid")
        return State.RUNNING if Process.is_alive(pid_path) else State.DEAD

    # will open a socket, send a request, read the response and close the socket.
    def _send_and_handle_socket_request(
        self, request: json_rpc.Request, version_hash: str
    ) -> None:
        try:
            stderr_file = os.path.join(self._log_directory, "server/server.stdout")
            with subprocess.Popen(
                ["tail", "--follow", "--lines=0", stderr_file],
                stdout=subprocess.PIPE,
                stderr=subprocess.DEVNULL,
                universal_newlines=True,
            ) as stderr_tail:
                try:
                    with SocketConnection(self._log_directory) as socket_connection:
                        socket_connection.perform_handshake(version_hash)
                        # pyre-fixme[6]: Expected `Iterable[str]` for 1st param but
                        #  got `Optional[IO[typing.Any]]`.
                        with StreamLogger(stderr_tail.stdout):
                            socket_connection.send(request)
                            response = socket_connection.read()
                        result = _convert_json_response_to_result(response)
                        result.check()
                        self._socket_result_handler(result)
                finally:
                    stderr_tail.terminate()
        except (
            SocketException,
            ResourceWarning,
            ClientException,
            json_rpc.JSONRPCException,
        ) as exception:
            LOG.error("Error while waiting for server: %s", str(exception))
            LOG.error("Run `pyre restart` in order to restart the server.")
            self._exit_code = ExitCode.FAILURE

    # Will be overwritten in subclasses to specialize how json socket
    # responses are handled.
    def _socket_result_handler(self, result: Result) -> None:
        log.stdout.write(result.output)

    def profiling_log_path(self) -> str:
        return os.path.join(self._log_directory, "profiling.log")

    @property
    def analysis_directory(self) -> AnalysisDirectory:
        return self._analysis_directory

    @property
    def configuration(self) -> Configuration:
        return self._configuration

    def _enable_logging_section(self, section: str) -> None:
        logging_sections = self._logging_sections
        if logging_sections:
            self._logging_sections = logging_sections + "," + section
        else:
            self._logging_sections = section



from types import ModuleType
from typing import Any, Dict, Optional

class _TempModule:
    mod_name: str = ...
    module: ModuleType = ...
    def __init__(self, mod_name): ...
    def __enter__(self): ...
    def __exit__(self, *args): ...

class _ModifiedArgv0:
    value: Any = ...
    def __init__(self, value): ...
    def __enter__(self): ...
    def __exit__(self, *args): ...

def run_module(
    mod_name: str, init_globals: Optional[Dict[str, Any]] = ..., run_name: Optional[str] = ..., alter_sys: bool = ...
): ...
def run_path(path_name: str, init_globals: Optional[Dict[str, Any]] = ..., run_name: str = ...): ...



# pyre-unsafe

import gettext as gettext_module
from typing import Any

def translation(language: Any) -> DjangoTranslation: ...

class DjangoTranslation(gettext_module.GNUTranslations):
    def __init__(self, language): ...
    def _new_gnu_trans(self, localedir, use_null_fallback=True): ...
    def _init_translation_catalog(self): ...
    def _add_installed_apps_translations(self): ...
    def _add_local_translations(self): ...
    def _add_fallback(self): ...
    def merge(self, other): ...
    def language(self): ...
    def to_language(self): ...



# pyre-unsafe

from typing import Any, ClassVar, Type

from django.core.exceptions import (
    MultipleObjectsReturned as MultipleObjectsReturnedException,
)
from django.db.aggregates import (
    Avg as Avg,
    Count as Count,
    Max as Max,
    Min as Min,
    StdDev as StdDev,
    Sum as Sum,
    Variance as Variance,
)
from django.db.models.deletion import (
    CASCADE,
    DO_NOTHING,
    PROTECT,
    RESTRICT,
    SET_DEFAULT,
    SET_NULL,
)
from django.db.models.enums import IntegerChoices, TextChoices
from django.db.models.fields import (
    AutoField,
    BigIntegerField,
    BooleanField,
    CharField,
    DateField,
    DateTimeField,
    DecimalField,
    EmailField,
    FloatField,
    IntegerField,
    NullBooleanField,
    PositiveIntegerField,
    PositiveSmallIntegerField,
    SmallAutoField,
    TextField,
    TimeField,
    URLField,
)
from django.db.models.fields.related import ForeignKey
from django.db.models.fields.subclassing import SubfieldBase
from django.db.models.manager import Manager
from django.db.models.options import Options

class Model:
    id: int = ...
    objects: Manager = ...
    _meta: Options = ...
    DoesNotExist: Any
    MultipleObjectsReturned: ClassVar[Type[MultipleObjectsReturnedException]] = ...
    def __init__(self, *args, **kwargs) -> None: ...
    def save(self) -> None: ...



# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from typing import TYPE_CHECKING, List

import graphene
from graphene import relay

from .models import Issue, IssueInstance, SharedText


if TYPE_CHECKING:
    from graphql.execution.base import ResolveInfo


if TYPE_CHECKING:
    from graphql.execution.base import ResolveInfo


SharedTextType = SharedText.generateSQLAlchemyObject()
IssueInstanceType = IssueInstance.generateSQLAlchemyObject()
IssueType = Issue.generateSQLAlchemyObject()


class IssueConnection(relay.Connection):
    class Meta:
        node = IssueInstanceType


class Query(graphene.ObjectType):
    node = relay.Node.Field()
    issue_instances = relay.ConnectionField(IssueConnection)

    def resolve_issue_instances(
        self, info: "ResolveInfo", **args
    ) -> List[IssueInstance]:
        issue_instances_query = IssueInstanceType.get_query(info)
        return issue_instances_query.all()


schema = graphene.Schema(query=Query, auto_camelcase=False)



from typing import Any, List, Optional, Sequence, TypeVar

_T = TypeVar("_T")

class StrongRandom:
    def __init__(self, rng: Optional[Any] = ..., randfunc: Optional[Any] = ...) -> None: ...
    def getrandbits(self, k: int) -> int: ...
    def randrange(self, start: int, stop: int = ..., step: int = ...) -> int: ...
    def randint(self, a: int, b: int) -> int: ...
    def choice(self, seq: Sequence[_T]) -> _T: ...
    def shuffle(self, x: Sequence[Any]): ...
    def sample(self, population: Sequence[_T], k: int) -> List[_T]: ...

def getrandbits(k: int) -> int: ...
def randrange(start: int, stop: int = ..., step: int = ...) -> int: ...
def randint(a: int, b: int) -> int: ...
def choice(seq: Sequence[_T]) -> _T: ...
def shuffle(x: Sequence[Any]): ...
def sample(population: Sequence[_T], k: int) -> List[_T]: ...



# pyre-unsafe

from typing import (
    Any,
    Callable,
    Generator,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

_CallableT = TypeVar("_CallableT", bound=Callable)

class Signal:
    send: Any = ...
    receivers: Any
    sender_receivers_cache: Any
    def __init__(
        self, providing_args: Optional[Iterable[Any]] = None, use_caching: bool = False
    ) -> None: ...
    def has_listeners(self, sender: Optional[Any] = None) -> bool: ...
    async def async_send(
        self, sender: Any, **named: Any
    ) -> List[Tuple[Callable[..., Any], Callable[..., Any]]]: ...
    def send(
        self, sender: Any, **named: Any
    ) -> List[Tuple[Callable[..., Any], Callable[..., Any]]]: ...
    def connect(
        self,
        receiver: Callable[..., None],
        sender: Optional[Signal] = None,
        weak: bool = True,
        dispatch_uid: Optional[str] = None,
    ) -> None: ...
    def _live_receivers(self, sender: Optional[Any] = None) -> Any: ...

def receiver(
    signal: Union[Signal, List[Signal], Tuple[Signal, ...]], **kwargs: Any
) -> Callable[[_CallableT], _CallableT]: ...



from typing import Any, NamedTuple, Optional, Text

class _URLTuple(NamedTuple):
    scheme: Any
    netloc: Any
    path: Any
    query: Any
    fragment: Any

class BaseURL(_URLTuple):
    def replace(self, **kwargs): ...
    @property
    def host(self): ...
    @property
    def ascii_host(self): ...
    @property
    def port(self): ...
    @property
    def auth(self): ...
    @property
    def username(self): ...
    @property
    def raw_username(self): ...
    @property
    def password(self): ...
    @property
    def raw_password(self): ...
    def decode_query(self, *args, **kwargs): ...
    def join(self, *args, **kwargs): ...
    def to_url(self): ...
    def decode_netloc(self): ...
    def to_uri_tuple(self): ...
    def to_iri_tuple(self): ...
    def get_file_location(self, pathformat: Optional[Any] = ...): ...

class URL(BaseURL):
    def encode_netloc(self): ...
    def encode(self, charset: Text = ..., errors: Text = ...): ...

class BytesURL(BaseURL):
    def encode_netloc(self): ...
    def decode(self, charset: Text = ..., errors: Text = ...): ...

def url_parse(url, scheme: Optional[Any] = ..., allow_fragments: bool = ...): ...
def url_quote(string, charset: Text = ..., errors: Text = ..., safe: str = ..., unsafe: str = ...): ...
def url_quote_plus(string, charset: Text = ..., errors: Text = ..., safe: str = ...): ...
def url_unparse(components): ...
def url_unquote(string, charset: Text = ..., errors: Text = ..., unsafe: str = ...): ...
def url_unquote_plus(s, charset: Text = ..., errors: Text = ...): ...
def url_fix(s, charset: Text = ...): ...
def uri_to_iri(uri, charset: Text = ..., errors: Text = ...): ...
def iri_to_uri(iri, charset: Text = ..., errors: Text = ..., safe_conversion: bool = ...): ...
def url_decode(
    s,
    charset: Text = ...,
    decode_keys: bool = ...,
    include_empty: bool = ...,
    errors: Text = ...,
    separator: str = ...,
    cls: Optional[Any] = ...,
): ...
def url_decode_stream(
    stream,
    charset: Text = ...,
    decode_keys: bool = ...,
    include_empty: bool = ...,
    errors: Text = ...,
    separator: str = ...,
    cls: Optional[Any] = ...,
    limit: Optional[Any] = ...,
    return_iterator: bool = ...,
): ...
def url_encode(
    obj, charset: Text = ..., encode_keys: bool = ..., sort: bool = ..., key: Optional[Any] = ..., separator: bytes = ...
): ...
def url_encode_stream(
    obj,
    stream: Optional[Any] = ...,
    charset: Text = ...,
    encode_keys: bool = ...,
    sort: bool = ...,
    key: Optional[Any] = ...,
    separator: bytes = ...,
): ...
def url_join(base, url, allow_fragments: bool = ...): ...

class Href:
    base: Any
    charset: Text
    sort: Any
    key: Any
    def __init__(self, base: str = ..., charset: Text = ..., sort: bool = ..., key: Optional[Any] = ...): ...
    def __getattr__(self, name): ...
    def __call__(self, *path, **query): ...



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# Enables navigating traces with sinks back to the source kinds that can reach it

import logging
from collections import defaultdict, deque
from typing import Dict, List, Set, Tuple

from .models import IssueInstance, SharedTextKind, TraceFrame, TraceKind
from .pipeline import PipelineStep, Summary
from .trace_graph import TraceGraph


log = logging.getLogger("sapp")

FrameID = int
SinkToSourceKindMap = Dict[int, Set[int]]
SinkToSourceKindDepthMap = Dict[int, Dict[int, int]]


class PropagateSourceKindsToSinks(  # pyre-fixme[13]
    PipelineStep[TraceGraph, TraceGraph]
):
    """For all issues propagate source kinds to all reachable frames leading to
    sinks."""

    def __init__(self) -> None:
        super().__init__()
        self.summary: Summary
        self.graph: TraceGraph
        # pyre-fixme[8]
        self.visited: Dict[FrameID, SinkToSourceKindDepthMap] = defaultdict(
            lambda: defaultdict(lambda: defaultdict(lambda: 999999))
        )

    def _subtract_kinds(
        self,
        depth: int,
        kind_map: SinkToSourceKindMap,
        to_remove: SinkToSourceKindDepthMap,
    ) -> SinkToSourceKindMap:
        result = {}
        for sink_kind, source_kinds in kind_map.items():
            if sink_kind in to_remove:
                source_depths = to_remove[sink_kind]
                source_kinds = {
                    source_kind
                    for source_kind in source_kinds
                    if source_depths.get(source_kind, depth + 1) > depth
                }
            if len(source_kinds) > 0:
                result[sink_kind] = source_kinds
        return result

    def _update_visited(
        self, frame_id: FrameID, depth: int, kind_map: SinkToSourceKindMap
    ) -> None:
        for sink_kind, source_kinds in kind_map.items():
            visited_frame = self.visited[frame_id]
            for source_kind in source_kinds:
                visited_frame[sink_kind][source_kind] = depth

    def _propagate_source_kinds(self, instance: IssueInstance):
        """Propagate the source kinds of this issue instance to all reachable sink
        traces."""
        graph = self.graph

        initial_frames = graph.get_issue_instance_trace_frames(instance)
        initial_source_frames = [
            frame for frame in initial_frames if frame.kind == TraceKind.POSTCONDITION
        ]
        initial_sink_frames = [
            frame for frame in initial_frames if frame.kind == TraceKind.PRECONDITION
        ]
        source_kind_list = [
            graph.get_trace_frame_leaf_ids_by_kind(frame, SharedTextKind.source)
            for frame in initial_source_frames
        ]
        if len(source_kind_list) == 0:
            return

        source_kinds = set.union(*source_kind_list)
        if len(source_kinds) == 0:
            return
        self._propagate_kinds_along_traces(initial_sink_frames, source_kinds)

    def _propagate_kinds_along_traces(
        self, start_frames: List[TraceFrame], new_kinds: Set[int]
    ) -> None:
        graph = self.graph

        queue = deque(
            [
                (
                    start_frame,
                    {
                        sink_id: new_kinds
                        for sink_id in graph.get_trace_frame_leaf_ids(start_frame)
                    },
                    0,
                )
                for start_frame in start_frames
            ]
        )
        while len(queue) > 0:
            frame, kind_map, depth = queue.popleft()
            if len(kind_map) == 0:
                continue

            frame_id = frame.id.local_id
            if frame_id in self.visited:
                kind_map = self._subtract_kinds(depth, kind_map, self.visited[frame_id])
                if len(kind_map) == 0:
                    continue

            self._update_visited(frame_id, depth, kind_map)

            next_frames = [
                (next_frame, graph.get_trace_frame_leaf_ids(frame))
                for next_frame in self.graph.get_trace_frames_from_caller(
                    frame.kind, frame.callee_id, frame.callee_port
                )
            ]
            queue.extend(
                (
                    frame,
                    {
                        sink: sources
                        for sink, sources in kind_map.items()
                        if sink in sink_kinds
                    },
                    depth + 1,
                )
                for (frame, sink_kinds) in next_frames
            )

    def run(self, input: TraceGraph, summary: Summary) -> Tuple[TraceGraph, Summary]:
        graph = input
        self.summary = summary
        self.graph = graph

        log.info("Propagating source kinds to sinks")

        for instance in graph.get_issue_instances():
            self._propagate_source_kinds(instance)

        # Create new source assocs based on the visited results
        source_count = 0
        trace_frame_count = 0
        for trace_frame_id, kind_map in self.visited.items():
            trace_frame_count += 1
            trace_frame = graph.get_trace_frame_from_id(trace_frame_id)
            for source_depths in kind_map.values():
                for source_kind, depth in source_depths.items():
                    graph.add_trace_frame_leaf_by_local_id_assoc(
                        trace_frame, source_kind, depth
                    )
                    source_count += 1

        log.info(
            f"Added {source_count} source kinds to {trace_frame_count} trace frames"
        )

        return graph, summary



from typing import Any, Mapping

class TemplateSyntaxError(Exception):
    pass

class Template: ...

class Context:
    def __init__(self, dict_: Mapping[str, Any]) -> None: ...

class Engine:
    def from_string(self, template_code: str) -> Engine: ...



# pyre-unsafe

from typing import List

from django.http import HttpRequest

DEBUG: int = ...
INFO: int = ...
SUCCESS: int = ...
WARNING: int = ...
ERROR: int = ...

def add_message(request: HttpRequest, int, str) -> None: ...
def get_messages(request: HttpRequest) -> List[str]: ...



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import functools
import json
import logging
from pathlib import Path
from typing import List, NamedTuple, Optional

from .. import log
from ..analysis_directory import AnalysisDirectory
from ..configuration import Configuration
from .command import JSON, Command, CommandArguments
from .stop import Stop


LOG: logging.Logger = logging.getLogger(__name__)


ROOT_PLACEHOLDER_NAME = "<root>"


class ServerDetails(NamedTuple):
    local_root: str
    pid: int
    server_pid_path: Path

    @staticmethod
    def _from_server_path(
        server_pid_path: Path, dot_pyre_root: Path
    ) -> "ServerDetails":
        return ServerDetails(
            pid=int(server_pid_path.read_text()),
            server_pid_path=server_pid_path,
            local_root=str(server_pid_path.relative_to(dot_pyre_root).parent.parent),
        )

    def is_root(self) -> bool:
        return self.local_root == "."

    @property
    def name(self) -> str:
        if self.is_root():
            return ROOT_PLACEHOLDER_NAME
        else:
            return self.local_root


class Servers(Command):
    NAME: str = "servers"

    def __init__(
        self,
        command_arguments: CommandArguments,
        original_directory: str,
        *,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
        subcommand: Optional[str],
    ) -> None:
        super(Servers, self).__init__(
            command_arguments, original_directory, configuration, analysis_directory
        )
        self._subcommand = subcommand

    @staticmethod
    def from_arguments(
        arguments: argparse.Namespace,
        original_directory: str,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
    ) -> "Servers":
        return Servers(
            CommandArguments.from_arguments(arguments),
            original_directory,
            configuration=configuration,
            analysis_directory=analysis_directory,
            subcommand=arguments.servers_subcommand,
        )

    @classmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> None:
        servers_parser = parser.add_parser(
            cls.NAME,
            epilog="""
            Command to manipulate multiple Pyre servers.
            """,
        )
        servers_parser.set_defaults(command=cls.from_arguments)
        subparsers = servers_parser.add_subparsers(dest="servers_subcommand")

        subparsers.add_parser("list", help="List running servers.")
        subparsers.add_parser("stop", help="Stop all running servers.")

    def is_root_server_running(self) -> bool:
        return any(server.is_root() for server in self._all_server_details())

    @staticmethod
    def _print_server_details(
        all_server_details: List[ServerDetails], output_format: str
    ) -> None:
        if output_format == JSON:
            server_details = [
                {"pid": details.pid, "name": details.name}
                for details in all_server_details
            ]
            log.stdout.write(json.dumps(server_details))
        else:
            if len(all_server_details) > 0:
                log.stdout.write(f"{'PID':<8}  project\n")
                for details in all_server_details:
                    log.stdout.write(f"{details.pid:<8}  {details.name}\n")
            else:
                log.stdout.write("No servers running")

    def _find_servers(self) -> List[Path]:
        return list(self._dot_pyre_directory.glob("**/server.pid"))

    def _all_server_details(self) -> List[ServerDetails]:
        return sorted(
            (
                ServerDetails._from_server_path(server_path, self._dot_pyre_directory)
                for server_path in self._find_servers()
            ),
            key=lambda details: details.local_root,
        )

    def _stop_servers(self, servers: List[ServerDetails]) -> None:
        for server in servers:
            LOG.warning("Stopping server for `%s` with pid %d", server.name, server.pid)
            Stop(
                command_arguments=self._command_arguments,
                original_directory=str(Path(self._project_root, server.local_root)),
            ).run()

    def _run(self) -> None:
        all_server_details = self._all_server_details()

        if self._subcommand == "list" or self._subcommand is None:
            self._print_server_details(all_server_details, self._output)
        elif self._subcommand == "stop":
            self._stop_servers(all_server_details)



# pyre-unsafe

from typing import Any, Dict, Optional

class SessionBase:
    session_key: Optional[str]
    _session_key: Optional[str]
    def __getitem__(self, key: str) -> Any: ...
    def __setitem__(self, key: str, value: Any) -> None: ...
    def get(self, key: str, default: Any = ...) -> Any: ...
    def update(self, dict_: Dict) -> None: ...
    def clear(self) -> None: ...



from django.http import HttpResponse, HttpResponseNotAllowed
from django.template.response import TemplateResponse
from django.utils.decorators import classonlymethod

logger = logging.getLogger("django.request")

class ContextMixin:
    extra_context = None
    def get_context_data(self, **kwargs): ...

class View:
    http_method_names = [
        "get",
        "post",
        "put",
        "patch",
        "delete",
        "head",
        "options",
        "trace",
    ]
    def __init__(self, **kwargs): ...
    @classonlymethod
    def as_view(cls, **initkwargs): ...
    def setup(self, request, *args, **kwargs): ...
    def dispatch(self, request, *args, **kwargs): ...
    def http_method_not_allowed(
        self, request, *args, **kwargs
    ) -> HttpResponseNotAllowed: ...
    def options(self, request, *args, **kwargs) -> HttpResponse: ...

class TemplateResponseMixin:
    template_name = None
    template_engine = None
    response_class = TemplateResponse
    content_type = None
    def render_to_response(self, context, **response_kwargs): ...
    def get_template_names(self): ...

class TemplateView(TemplateResponseMixin, ContextMixin, View):
    def get(self, request, *args, **kwargs) -> HttpResponse: ...

class RedirectView(View):
    permanent = False
    url = None
    pattern_name = None
    query_string = False
    def get_redirect_url(self, *args, **kwargs) -> str: ...
    def get(self, request, *args, **kwargs) -> HttpResponse: ...
    def head(self, request, *args, **kwargs) -> HttpResponse: ...
    def post(self, request, *args, **kwargs) -> HttpResponse: ...
    def options(self, request, *args, **kwargs) -> HttpResponse: ...
    def delete(self, request, *args, **kwargs) -> HttpResponse: ...
    def put(self, request, *args, **kwargs) -> HttpResponse: ...
    def patch(self, request, *args, **kwargs) -> HttpResponse: ...



# pyre-unsafe

from email.mime.base import MIMEBase
from typing import Any, Iterable, Mapping, Optional, Sequence, Tuple, Union

from django.core.mail.backends.base import BaseEmailBackend

class SafeMIMEText: ...
class SafeMIMEMultipart: ...

class EmailMessage:
    def __init__(
        self,
        subject: str,
        body: str,
        from_email: Optional[str],
        to: Optional[Iterable[str]],
        bcc: Optional[Iterable[str]],
        connection: Any,
        attachments: Union[MIMEBase, Tuple[str, str, str]],
        headers: Mapping[str, str],
        cc: Optional[Iterable[str]],
        reply_to: Optional[Iterable[str]],
    ) -> None: ...
    def get_connection(self, fail_silently: bool = False) -> BaseEmailBackend: ...
    def message(self) -> Union[SafeMIMEText, SafeMIMEMultipart]: ...
    def recipients(self) -> Sequence[str]: ...
    def send(self, fail_silently: bool = False) -> int: ...
    # def attach(self, filename=None, content=None, mimetype=None): ...
    def attach_file(self, path: str, mimetype: Optional[str] = None) -> None: ...



# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import logging
import os
from typing import Optional

import sqlalchemy
from flask import Flask, send_from_directory
from flask_cors import CORS
from flask_graphql import GraphQLView
from sqlalchemy.orm import Session, scoped_session, sessionmaker

from .models import Base
from .schema import schema


application = Flask(
    __name__,
    static_folder=os.path.join(os.path.dirname(__file__), "..", "frontend", "build"),
)
# TODO(T58070180): we should remove the cross origin policy.
CORS(application)
application.debug = True

session: Optional[Session] = None


@application.teardown_appcontext
def shutdown_session(exception=None):
    if session is not None:
        session.remove()


@application.route("/", defaults={"path": ""})
@application.route("/<path:path>")
def serve(path):
    if path != "" and os.path.exists(application.static_folder + "/" + path):
        return send_from_directory(application.static_folder, path)
    else:
        logging.error(application.static_folder)
        return send_from_directory(application.static_folder, "index.html")


def start_app(database):
    application.add_url_rule(
        "/graphql",
        view_func=GraphQLView.as_view("graphql", schema=schema, graphiql=True),
    )
    engine = sqlalchemy.create_engine(
        sqlalchemy.engine.url.URL("sqlite", database=database.dbname),
        echo=False,
        poolclass=None,
    )
    session = scoped_session(sessionmaker(bind=engine))
    Base.query = session.query_property()

    application.run()



# pyre-unsafe

from typing import Any, Optional

from django.http.request import HttpRequest
from django.http.response import HttpResponseRedirect

def get_object_or_404(klass: Any, *args: Any, **kwargs: Any) -> Any: ...
def redirect(to: Any, *args: Any, **kwargs: Any) -> HttpResponseRedirect: ...
def render(
    request: HttpRequest,
    template_name: str,
    context=None,
    content_type: Any = ...,
    status: Optional[int] = ...,
    using: Any = ...,
): ...



from enum import Enum
from typing import Sequence, Tuple

class Choices(Enum):
    def __str__(self) -> str: ...

class IntegerChoices(int, Choices):
    def choices(self) -> Sequence[Tuple[int, str]]: ...

class TextChoices(str, Choices):
    @property
    def choices(self) -> Sequence[Tuple[str, str]]: ...



# Copyright (c) 2019-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import functools
import os
from pathlib import Path
from typing import Any, Dict, List, Sequence, Set

from . import json_rpc, watchman
from .analysis_directory import AnalysisDirectory
from .buck import BuckException
from .configuration import Configuration
from .filesystem import find_root
from .process import Process
from .socket_connection import SocketConnection

# We use the `LOG` from watchman due to its better formatting in log files
from .watchman import LOG, Subscriber, Subscription


class MonitorException(Exception):
    pass


def _log_paths(message: str, paths: Sequence[str]) -> None:
    path_count = len(paths)
    log_threshold = 30
    if path_count <= log_threshold:
        LOG.info(f"{message} {paths}")
    else:
        additional_count = path_count - log_threshold
        LOG.info(f"{message} {paths[:log_threshold]} (and {additional_count} more)")


class ProjectFilesMonitor(Subscriber):
    """
        Logs from this monitor are found in
        .pyre/<local root>/file_monitor/file_monitor.log
        One file monitor is spawned per pyre server. When a server is stopped,
        the process of pyre file monitor associated to it is killed.
    """

    NAME = "file_monitor"

    def __init__(
        self,
        configuration: Configuration,
        project_root: str,
        analysis_directory: AnalysisDirectory,
    ) -> None:
        super(ProjectFilesMonitor, self).__init__(self.base_path(configuration))
        self._configuration = configuration
        self._analysis_directory = analysis_directory

        self._extensions: Set[str] = set(
            ["py", "pyi", "thrift"] + configuration.extensions
        )

        self._watchman_path: str = self._find_watchman_path(project_root)

        self.socket_connection = SocketConnection(self._configuration.log_directory)
        self.socket_connection.connect()
        self.socket_connection.perform_handshake(self._configuration.version_hash)

    @property
    def _name(self) -> str:
        return self.NAME

    @property
    @functools.lru_cache(1)
    def _subscriptions(self) -> List[Subscription]:
        subscription = {
            "empty_on_fresh_instance": True,
            "expression": [
                "allof",
                ["type", "f"],
                [
                    "anyof",
                    *[["suffix", extension] for extension in self._extensions],
                    ["match", "TARGETS"],
                ],
            ],
            "fields": ["name"],
        }
        return [
            Subscription(
                self._watchman_path, "pyre_file_change_subscription", subscription
            )
        ]

    @staticmethod
    def base_path(configuration: Configuration) -> str:
        return os.path.join(configuration.log_directory, ProjectFilesMonitor.NAME)

    def _handle_response(self, response: Dict[str, Any]) -> None:
        try:
            absolute_paths = [
                os.path.join(response["root"], path) for path in response["files"]
            ]
            _log_paths("Received Watchman update for files", absolute_paths)

            updated_paths = self._analysis_directory.process_updated_files(
                absolute_paths
            )

            if updated_paths.is_empty():
                LOG.info("Skipping update: Pyre doesn't track any of these files.")
                return

            _log_paths(
                "Notifying server of update to files", updated_paths.updated_paths
            )
            _log_paths("  and invalidation of files", updated_paths.deleted_paths)

            message = json_rpc.Request(
                method="updateFiles",
                parameters={
                    "files": updated_paths.updated_paths,
                    "invalidated": updated_paths.deleted_paths,
                },
            )
            if not message.write(self.socket_connection.output):
                LOG.info("Failed to communicate with server. Shutting down.")
                self._alive = False  # terminate daemon
                self.socket_connection.close()

        except KeyError:
            pass

        except BuckException:
            LOG.info("Unable to build project.")
            pass

        except Exception as exception:
            LOG.info(f"Exception during handling of file update: {exception}")
            raise exception

    @staticmethod
    def _find_watchman_path(directory: str) -> str:
        watchman_path = find_root(directory, ".watchmanconfig")
        if not watchman_path:
            raise MonitorException(
                "Could not find a watchman directory from "
                "the current directory `{}`".format(directory)
            )
        return watchman_path

    @staticmethod
    def restart_if_dead(
        configuration: Configuration,
        project_root: str,
        analysis_directory: AnalysisDirectory,
    ) -> None:
        pid_path = watchman.compute_pid_path(
            ProjectFilesMonitor.base_path(configuration), ProjectFilesMonitor.NAME
        )
        if Process.is_alive(Path(pid_path)):
            return
        LOG.debug("File monitor is not running.")
        try:
            ProjectFilesMonitor(
                configuration, project_root, analysis_directory
            ).daemonize()
            LOG.debug("Restarted file monitor.")
        except MonitorException as exception:
            LOG.warning(f"Failed to restart file monitor: {exception}")



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import argparse
import os
from typing import List, Optional

from typing_extensions import Final

from .. import log
from ..analysis_directory import AnalysisDirectory, resolve_analysis_directory
from ..configuration import Configuration
from ..filesystem import readable_directory, writable_directory
from .check import Check
from .command import CommandArguments


class Analyze(Check):
    NAME = "analyze"

    def __init__(
        self,
        command_arguments: CommandArguments,
        original_directory: str,
        *,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
        analysis: str,
        taint_models_path: List[str],
        no_verify: bool,
        save_results_to: Optional[str],
        dump_call_graph: bool,
        repository_root: Optional[str],
        rules: Optional[List[int]],
        find_obscure_flows: bool = False,
    ) -> None:
        super(Analyze, self).__init__(
            command_arguments,
            original_directory,
            configuration=configuration,
            analysis_directory=analysis_directory,
        )
        self._analysis: str = analysis
        self._taint_models_path: List[str] = (
            taint_models_path or self._configuration.taint_models_path
        )
        self._no_verify: bool = no_verify
        self._save_results_to: Final[Optional[str]] = save_results_to
        self._dump_call_graph: bool = dump_call_graph
        self._repository_root: Final[Optional[str]] = repository_root
        self._rules: Final[Optional[List[int]]] = rules
        self._find_obscure_flows: bool = find_obscure_flows

    @staticmethod
    def from_arguments(
        arguments: argparse.Namespace,
        original_directory: str,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
    ) -> "Analyze":
        return Analyze(
            CommandArguments.from_arguments(arguments),
            original_directory,
            configuration=configuration,
            analysis_directory=analysis_directory,
            analysis=arguments.analysis,
            taint_models_path=arguments.taint_models_path,
            no_verify=arguments.no_verify,
            save_results_to=arguments.save_results_to,
            dump_call_graph=arguments.dump_call_graph,
            repository_root=arguments.repository_root,
            rules=arguments.rule,
            find_obscure_flows=arguments.find_obscure_flows,
        )

    @classmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> None:
        analyze = parser.add_parser(cls.NAME)
        analyze.set_defaults(command=cls.from_arguments)
        analyze.add_argument(
            "analysis",
            nargs="?",
            default="taint",
            help="Type of analysis to run: {taint}",
        )
        analyze.add_argument(
            "--taint-models-path",
            action="append",
            default=[],
            type=readable_directory,
            help="Location of taint models",
        )
        analyze.add_argument(
            "--no-verify",
            action="store_true",
            help="Do not verify models for the taint analysis.",
        )
        analyze.add_argument(
            "--save-results-to",
            default=None,
            type=writable_directory,
            help="Directory to write analysis results to.",
        )
        analyze.add_argument("--dump-call-graph", action="store_true")
        analyze.add_argument("--repository-root", type=os.path.abspath)
        analyze.add_argument("--rule", action="append", type=int)
        analyze.add_argument(
            "--find-obscure-flows",
            action="store_true",
            help="Perform a taint analysis to find flows through obscure models.",
        )

    def generate_analysis_directory(self) -> AnalysisDirectory:
        return resolve_analysis_directory(
            self._source_directories,
            self._targets,
            self._configuration,
            self._original_directory,
            self._project_root,
            filter_directory=self._filter_directory,
            use_buck_builder=self._use_buck_builder,
            debug=self._debug,
            buck_mode=self._buck_mode,
            isolate=True,
        )

    def _flags(self) -> List[str]:
        flags = super()._flags()
        flags.extend(["-analysis", self._analysis])
        if self._taint_models_path:
            for path in self._taint_models_path:
                flags.extend(["-taint-models", path])
        save_results_to = self._save_results_to
        if save_results_to:
            flags.extend(["-save-results-to", save_results_to])
        if self._dump_call_graph:
            flags.append("-dump-call-graph")
        if self._no_verify:
            flags.append("-no-verify")
        repository_root = self._repository_root
        if repository_root:
            flags.extend(["-repository-root", repository_root])
        rules = self._rules
        if rules is not None:
            flags.extend(["-rules", ",".join(str(rule) for rule in rules)])
        if self._find_obscure_flows:
            flags.append("-find-obscure-flows")
        return flags

    def _run(self, retries: int = 1) -> None:
        self._analysis_directory.prepare()
        result = self._call_client(command=self.NAME)
        result.check()
        if self._save_results_to is None:
            log.stdout.write(result.output)



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections import defaultdict
from typing import DefaultDict, Dict, Iterable, List, Optional, Set, Tuple

from .bulk_saver import BulkSaver
from .models import (
    DBID,
    Issue,
    IssueInstance,
    IssueInstanceFixInfo,
    SharedText,
    SharedTextKind,
    TraceFrame,
    TraceFrameAnnotation,
    TraceKind,
)


class TraceGraph(object):
    """Represents a graph of the Zoncolan trace steps. Nodes of the graph are
    the issues, preconditions, postconditions, sources and sinks. Edges are
    the the assocs, and for pre/postconditions, the map of 'caller->callee'
    gives a one direction edge to the next pre/postcondition, and the map of
    'callee->caller' gives the reverse edge.
    """

    def __init__(self) -> None:
        self._issues: Dict[int, Issue] = {}
        self._issue_instances: Dict[int, IssueInstance] = {}
        self._trace_annotations: Dict[int, TraceFrameAnnotation] = {}

        # Create a mapping of (caller_id, caller_port) to the corresponding
        # trace frame's id.
        self._trace_frames_map: DefaultDict[
            TraceKind, DefaultDict[Tuple[int, str], Set[int]]
        ] = defaultdict(lambda: defaultdict(set))

        # Similar to _trace_frames_map, but maps the reverse direction
        # of the trace graph, i.e. (callee_id, callee_port) to the
        # trace_frame_id.
        self._trace_frames_rev_map: DefaultDict[
            TraceKind, DefaultDict[Tuple[int, str], Set[int]]
        ] = defaultdict(lambda: defaultdict(set))

        self._trace_frames: Dict[int, TraceFrame] = {}

        self._shared_texts: Dict[int, SharedText] = {}
        self._shared_text_lookup: (
            DefaultDict[SharedTextKind, Dict[str, int]]
        ) = defaultdict(dict)

        self._trace_frame_leaf_assoc: DefaultDict[
            int, Set[Tuple[int, int]]
        ] = defaultdict(set)

        self._trace_frame_issue_instance_assoc: DefaultDict[
            int, Set[int]
        ] = defaultdict(set)
        self._issue_instance_trace_frame_assoc: DefaultDict[
            int, Set[int]
        ] = defaultdict(set)

        self._trace_frame_annotation_trace_frame_assoc: DefaultDict[
            int, Set[int]
        ] = defaultdict(set)
        self._trace_frame_trace_frame_annotation_assoc: DefaultDict[
            int, Set[int]
        ] = defaultdict(set)

        self._issue_instance_shared_text_assoc: DefaultDict[
            int, Set[int]
        ] = defaultdict(set)
        self._shared_text_issue_instance_assoc: DefaultDict[
            int, Set[int]
        ] = defaultdict(set)

        self._issue_instance_fix_info: Dict[int, IssueInstanceFixInfo] = {}

        # !!!!! IMPORTANT !!!!!
        # IF YOU ARE ADDING MORE FIELDS/EDGES TO THIS GRAPH, CHECK IF
        # TrimmedTraceGraph NEEDS TO BE UPDATED AS WELL.
        #
        # TrimmedTraceGraph will populate itself from this object. It searches
        # TrimmedGraph for nodes and edges of 'affected_files' and copies them
        # over. If new fields/edges are added, these may need to be copied in
        # TrimmedTraceGraph as well.

    def add_issue(self, issue: Issue) -> None:
        assert issue.id.local_id not in self._issues, "Issue already exists"
        self._issues[issue.id.local_id] = issue

    def get_issue(self, issue_id: DBID) -> Issue:
        return self._issues[issue_id.local_id]

    def add_issue_instance(self, instance: IssueInstance) -> None:
        assert (
            instance.id.local_id not in self._issue_instances
        ), "Instance already exists"
        self._issue_instances[instance.id.local_id] = instance

    def get_issue_instances(self) -> Iterable[IssueInstance]:
        return (instance for instance in self._issue_instances.values())

    def add_issue_instance_fix_info(
        self, instance: IssueInstance, fix_info: IssueInstanceFixInfo
    ) -> None:
        assert (
            instance.id.local_id not in self._issue_instance_fix_info
        ), "Instance fix info already exists"
        self._issue_instance_fix_info[instance.id.local_id] = fix_info

    def get_text(self, shared_text_id: DBID) -> str:
        return self._shared_texts[shared_text_id.local_id].contents

    def get_shared_text_by_local_id(self, shared_text_id: int) -> SharedText:
        return self._shared_texts[shared_text_id]

    def get_shared_text(
        self, kind: SharedTextKind, content: str
    ) -> Optional[SharedText]:
        if kind in self._shared_text_lookup:
            contents = self._shared_text_lookup[kind]
            if content in contents and contents[content] in self._shared_texts:
                return self._shared_texts[contents[content]]
        return None

    def has_trace_frames_with_caller(
        self, kind: TraceKind, caller_id: DBID, caller_port: str
    ) -> bool:
        if self._trace_frames_map[kind]:
            key = (caller_id.local_id, caller_port)
            return key in self._trace_frames_map[kind]
        else:
            return False

    def has_postconditions_with_caller(self, caller_id: DBID, caller_port: str) -> bool:
        return self.has_trace_frames_with_caller(
            TraceKind.postcondition, caller_id, caller_port
        )

    def has_preconditions_with_caller(self, caller_id: DBID, caller_port: str) -> bool:
        return self.has_trace_frames_with_caller(
            TraceKind.precondition, caller_id, caller_port
        )

    def add_trace_annotation(self, annotation: TraceFrameAnnotation) -> None:
        self._trace_annotations[annotation.id.local_id] = annotation

    def get_condition_annotations(self, cond_id: int) -> List[TraceFrameAnnotation]:
        return [
            t
            for t in self._trace_annotations.values()
            if t.trace_frame_id.local_id == cond_id
        ]

    def get_annotation_trace_frames(self, ann_id: int) -> List[TraceFrame]:
        if ann_id in self._trace_frame_annotation_trace_frame_assoc:
            return [
                self.get_trace_frame_from_id(tf_id)
                for tf_id in self._trace_frame_annotation_trace_frame_assoc[ann_id]
            ]
        else:
            return []

    def add_trace_frame(self, trace_frame: TraceFrame) -> None:
        key = (trace_frame.caller_id.local_id, trace_frame.caller_port)
        rev_key = (trace_frame.callee_id.local_id, trace_frame.callee_port)
        self._trace_frames_map[trace_frame.kind][key].add(trace_frame.id.local_id)
        self._trace_frames_rev_map[trace_frame.kind][rev_key].add(
            trace_frame.id.local_id
        )
        self._trace_frames[trace_frame.id.local_id] = trace_frame

    def get_trace_frames_from_caller(
        self, kind: TraceKind, caller_id: DBID, caller_port: str
    ) -> List[TraceFrame]:
        key = (caller_id.local_id, caller_port)
        return [
            self._trace_frames[trace_frame_id]
            for trace_frame_id in self._trace_frames_map[kind][key]
        ]

    def get_trace_frame_from_id(self, id: int) -> TraceFrame:
        return self._trace_frames[id]

    def add_shared_text(self, shared_text: SharedText) -> None:
        assert (
            shared_text.id.local_id not in self._shared_texts
        ), "Shared text already exists"
        assert (
            shared_text.kind not in self._shared_text_lookup
            or shared_text.contents not in self._shared_text_lookup[shared_text.kind]
        ), "Shared text with same kind, contents exists"

        self._shared_texts[shared_text.id.local_id] = shared_text

        # Allow look up of SharedTexts by name and kind (to optimize
        # get_shared_text which is called when parsing each issue instance)
        self._shared_text_lookup[shared_text.kind][
            shared_text.contents
        ] = shared_text.id.local_id

    def add_trace_frame_leaf_assoc(
        self, trace_frame: TraceFrame, leaf: SharedText, depth: int
    ) -> None:
        self._trace_frame_leaf_assoc[trace_frame.id.local_id].add(
            (leaf.id.local_id, depth)
        )

    def add_trace_frame_leaf_by_local_id_assoc(
        self, trace_frame: TraceFrame, leaf_id: int, depth: int
    ) -> None:
        self._trace_frame_leaf_assoc[trace_frame.id.local_id].add((leaf_id, depth))

    def get_trace_frame_leaf_ids(self, trace_frame: TraceFrame) -> Set[int]:
        ids: Set[int] = {
            id for (id, depth) in self._trace_frame_leaf_assoc[trace_frame.id.local_id]
        }
        return ids

    def get_trace_frame_leaf_ids_by_kind(
        self, trace_frame: TraceFrame, kind: SharedTextKind
    ) -> Set[int]:
        return {
            id
            for (id, depth) in self._trace_frame_leaf_assoc[trace_frame.id.local_id]
            if self._shared_texts[id].kind == kind
        }

    def get_trace_frame_leaf_ids_with_depths(
        self, trace_frame: TraceFrame
    ) -> Set[Tuple[int, int]]:
        return self._trace_frame_leaf_assoc[trace_frame.id.local_id]

    def add_issue_instance_trace_frame_assoc(
        self, instance: IssueInstance, trace_frame: TraceFrame
    ) -> None:
        self._issue_instance_trace_frame_assoc[instance.id.local_id].add(
            trace_frame.id.local_id
        )
        self._trace_frame_issue_instance_assoc[trace_frame.id.local_id].add(
            instance.id.local_id
        )

    def add_trace_frame_annotation_trace_frame_assoc(
        self, annotation: TraceFrameAnnotation, trace_frame: TraceFrame
    ) -> None:
        self._trace_frame_annotation_trace_frame_assoc[annotation.id.local_id].add(
            trace_frame.id.local_id
        )
        self._trace_frame_trace_frame_annotation_assoc[trace_frame.id.local_id].add(
            annotation.id.local_id
        )

    def get_issue_instance_trace_frames(
        self, instance: IssueInstance
    ) -> List[TraceFrame]:
        if instance.id.local_id in self._issue_instance_trace_frame_assoc:
            return [
                self.get_trace_frame_from_id(id)
                for id in self._issue_instance_trace_frame_assoc[instance.id.local_id]
            ]
        else:
            return []

    def get_next_trace_frames(self, trace_frame: TraceFrame) -> Iterable[TraceFrame]:
        return self.get_trace_frames_from_caller(
            trace_frame.kind, trace_frame.callee_id, trace_frame.callee_port
        )

    def add_issue_instance_shared_text_assoc(
        self, instance: IssueInstance, shared_text: SharedText
    ) -> None:
        self._issue_instance_shared_text_assoc[instance.id.local_id].add(
            shared_text.id.local_id
        )
        self._shared_text_issue_instance_assoc[shared_text.id.local_id].add(
            instance.id.local_id
        )

    def get_issue_instance_shared_texts(
        self, instance_id: int, kind: SharedTextKind
    ) -> List[SharedText]:
        return [
            self._shared_texts[msg_id]
            for msg_id in self._issue_instance_shared_text_assoc[instance_id]
            if self._shared_texts[msg_id].kind == kind
        ]

    def update_bulk_saver(self, bulk_saver: BulkSaver) -> None:
        bulk_saver.add_all(list(self._issues.values()))
        bulk_saver.add_all(list(self._issue_instances.values()))
        bulk_saver.add_all(list(self._trace_frames.values()))
        bulk_saver.add_all(list(self._issue_instance_fix_info.values()))
        bulk_saver.add_all(list(self._trace_annotations.values()))
        bulk_saver.add_all(list(self._shared_texts.values()))

        self._save_issue_instance_trace_frame_assoc(bulk_saver)
        self._save_trace_frame_leaf_assoc(bulk_saver)
        self._save_issue_instance_shared_text_assoc(bulk_saver)
        self._save_trace_frame_annotation_trace_frame_assoc(bulk_saver)

    def _save_issue_instance_trace_frame_assoc(self, bulk_saver: BulkSaver) -> None:
        for (
            trace_frame_id,
            instance_ids,
        ) in self._trace_frame_issue_instance_assoc.items():
            for instance_id in instance_ids:
                bulk_saver.add_issue_instance_trace_frame_assoc(
                    self._issue_instances[instance_id],
                    self._trace_frames[trace_frame_id],
                )

    def _save_trace_frame_annotation_trace_frame_assoc(
        self, bulk_saver: BulkSaver
    ) -> None:
        for (
            trace_annotation_id,
            trace_frame_ids,
        ) in self._trace_frame_annotation_trace_frame_assoc.items():
            for trace_frame_id in trace_frame_ids:
                bulk_saver.add_trace_frame_annotation_trace_frame_assoc(
                    self._trace_annotations[trace_annotation_id],
                    self._trace_frames[trace_frame_id],
                )

    def _save_trace_frame_leaf_assoc(self, bulk_saver: BulkSaver) -> None:
        for trace_frame_id, leaf_ids in self._trace_frame_leaf_assoc.items():
            for (leaf_id, depth) in leaf_ids:
                bulk_saver.add_trace_frame_leaf_assoc(
                    self._shared_texts[leaf_id],
                    self._trace_frames[trace_frame_id],
                    depth,
                )

    def _save_issue_instance_shared_text_assoc(self, bulk_saver: BulkSaver) -> None:
        for (
            shared_text_id,
            instance_ids,
        ) in self._shared_text_issue_instance_assoc.items():
            for instance_id in instance_ids:
                bulk_saver.add_issue_instance_shared_text_assoc(
                    self._issue_instances[instance_id],
                    self._shared_texts[shared_text_id],
                )



#!/usr/bin/env python3

# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging

import click

from .cli_lib import commands, common_options
from .context import Context
from .db import DB, DBType
from .lint import lint
from .pysa_taint_parser import Parser


logger = logging.getLogger("sapp")


@common_options
@click.option(
    "--database-engine",
    "--database",
    type=click.Choice([DBType.SQLITE, DBType.MEMORY]),
    default=DBType.SQLITE,
    help="database engine to use",
)
@click.pass_context
def cli(
    ctx: click.Context, repository: str, database_name: str, database_engine: DBType
):
    ctx.obj = Context(
        repository=repository,
        database=DB(database_engine, database_name, assertions=True),
        parser_class=Parser,
    )
    logger.debug(f"Context: {ctx.obj}")


for command in commands:
    cli.add_command(command)
cli.add_command(lint)

if __name__ == "__main__":
    cli()



# pyre-unsafe

from typing import Any, Optional

from django.contrib.auth.base_user import AbstractBaseUser, BaseUserManager
from django.db import models

class UserManager(BaseUserManager):
    pass

class AnonymousUser(object):
    id: Optional[int]
    is_active: bool
    def is_anonymous(self) -> bool: ...
    def is_authenticated(self) -> bool: ...

class AbstractUser(AbstractBaseUser):
    pass

class User(AbstractUser):
    pass

def _user_has_module_perms(user: Any, app_label: Any) -> bool: ...



# pyre-unsafe

class CursorWrapper:
    def __init__(self, cursor, db): ...
    def __enter__(self) -> "CursorWrapper": ...
    def execute(self, sql, params=None): ...
    def executemany(self, sql, param_list): ...
    def fetchone(self): ...



# Stub for termcolor: https://pypi.python.org/pypi/termcolor
from typing import Any, Iterable, Optional, Text

def colored(
    text: Text, color: Optional[Text] = ..., on_color: Optional[Text] = ..., attrs: Optional[Iterable[Text]] = ...,
) -> Text: ...
def cprint(
    text: Text, color: Optional[Text] = ..., on_color: Optional[Text] = ..., attrs: Optional[Iterable[Text]] = ..., **kwargs: Any,
) -> None: ...



# pyre-unsafe

from typing import Any, Mapping, Optional

def int_to_base36(i: int) -> str: ...
def base36_to_int(s: str) -> int: ...
def urlencode(data: Mapping[str, Any], doseq: bool = ...) -> str: ...
def urlquote(url: str, safe: str = ...) -> str: ...
def http_date(epoch_seconds: Optional[float] = None) -> str: ...
def same_origin(lhs: str, rhs: str) -> bool: ...
def is_safe_url(url: str, host: Optional[str] = None) -> bool: ...



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import json
import os
from glob import glob
from pathlib import Path
from typing import IO, Any, Dict, Iterable, NamedTuple, Optional

from .sharded_files import ShardedFile


METADATA_GLOB = "*metadata.json"


# pyre-fixme[2]: Parameter annotation cannot contain `Any`.
class Metadata(NamedTuple):
    analysis_root: str
    repo_root: Optional[str] = None
    repository_name: Optional[str] = None
    tool: Optional[str] = None
    analysis_tool_version: Optional[str] = None
    commit_hash: Optional[str] = None
    job_instance: Optional[int] = None
    project: Optional[str] = None
    # Mapping from code to rule metadata.
    # pyre-ignore: we don't have a shape for rules yet.
    rules: Dict[int, Any] = {}

    @property
    def root(self) -> str:
        return self.repo_root or self.analysis_root


class AnalysisOutputError(Exception):
    pass


class AnalysisOutput(object):
    """Represents one of various ways the analysis output can be specified.

    Use "filename_spec" to represent either:
      A file name, a file handle, or a sharded file pattern.

    Use "filename_glob" to specify a set of filename patterns instead. This
    assumes the output lives in the given directory. Avoid patterns like '*'
    which will include the metadata.json file in the directory.

    Note that "filename_spec" has precedence over "filename_glob".

    Access to the output is provided via generators that provide file handles
    to the diagnostics json (issues), or the summary json (pre and post).
    """

    def __init__(
        self,
        *,
        directory: Optional[str] = None,
        filename_spec: Optional[str] = None,
        filename_glob: Optional[str] = None,
        file_handle: Optional[IO[str]] = None,
        metadata: Optional[Metadata] = None,
        tool: Optional[str] = None,
    ) -> None:
        self.directory = directory
        self.filename_spec = filename_spec
        self.filename_glob = filename_glob
        self.file_handle = file_handle
        self.metadata = metadata
        self.tool = tool

        if not filename_spec and file_handle and hasattr(file_handle, "name"):
            self.filename_spec = file_handle.name

    def __str__(self) -> str:
        if self.directory:
            return f"AnalysisOutput({repr(self.directory)})"

        return f"AnalysisOutput({repr(self.filename_spec)})"

    @classmethod
    def from_str(cls, identifier: str) -> "AnalysisOutput":
        if os.path.isdir(identifier):
            return cls.from_directory(identifier)
        elif os.path.isfile(identifier):
            return cls.from_file(identifier)
        elif os.path.isdir(os.path.dirname(identifier)) and "@" in os.path.basename(
            identifier
        ):
            return cls.from_file(identifier)
        else:
            raise AnalysisOutputError(f"Unrecognized identifier `{identifier}`")

    @classmethod
    def from_directory(cls, directory: str) -> "AnalysisOutput":
        metadata = {}
        for file in glob(os.path.join(directory, METADATA_GLOB)):
            with open(file) as f:
                metadata.update(json.load(f))

        # Note: filename_spec takes precedence over filename_glob.
        filename_spec = None
        filename_glob = None
        if "filename_spec" in metadata:
            filename_spec = os.path.join(
                directory, os.path.basename(metadata["filename_spec"])
            )
        elif "filename_glob" in metadata:
            filename_glob = metadata["filename_glob"]
            if not filename_glob:
                raise AnalysisOutputError(
                    f"Empty 'filename_glob' not allowed. In {METADATA_GLOB}, "
                    "Use either 'filename_spec' or specify something in "
                    "'filename_glob'."
                )
        else:
            # Legacy
            filename_spec = os.path.join(
                directory, os.path.basename(metadata["filenames"][0])
            )

        repo_root = metadata.get("repo_root")
        analysis_root = metadata["root"]

        rules = {rule["code"]: rule for rule in metadata.get("rules", [])}

        return cls(
            directory=directory,
            filename_spec=filename_spec,
            filename_glob=filename_glob,
            metadata=Metadata(
                analysis_tool_version=metadata["version"],
                commit_hash=metadata.get("commit"),
                analysis_root=analysis_root,
                repo_root=repo_root,
                job_instance=metadata.get("job_instance"),
                tool=metadata.get("tool"),
                repository_name=metadata.get("repository_name"),
                project=metadata.get("project"),
                rules=rules,
            ),
        )

    @classmethod
    def from_file(cls, file_name: str) -> "AnalysisOutput":
        # """Pass in either a single file name or a sharded file pattern.
        # Performs early validation by 1) opening the file if it is a single file,
        # or 2) computing and checking the file shards.
        # """
        return cls(filename_spec=file_name)

    @classmethod
    def from_handle(cls, file_handle: IO[str]) -> "AnalysisOutput":
        return cls(file_handle=file_handle)

    def file_handles(self) -> Iterable[IO[str]]:
        """Generates all file handles represented by the analysis.
        Callee owns file handle and closes it when the next is yielded or the
        generator ends.
        """
        if self.file_handle:
            # pyre-fixme[7]: Expected `Iterable[IO[str]]` but got
            #  `Generator[Optional[IO[str]], None, None]`.
            yield self.file_handle
            # pyre-fixme[16]: `Optional` has no attribute `close`.
            self.file_handle.close()
            self.file_handle = None
        else:
            for name in self.file_names():
                with open(name, "r") as f:
                    yield f

    def file_names(self) -> Iterable[str]:
        """Generates all file names that are used to generate file_handles.
        """
        filename_spec = self.filename_spec
        filename_glob = self.filename_glob
        if self.is_sharded():
            yield from ShardedFile(self.filename_spec).get_filenames()
        elif filename_glob is not None:
            directory = self.directory
            assert directory is not None
            # str() cast to convert the returned Path to string for a
            # consistent return type.
            for path in Path(directory).glob(filename_glob):
                yield str(path)
        elif filename_spec:
            yield filename_spec

    def is_sharded(self) -> bool:
        if self.filename_spec:
            # pyre-fixme[16]: `Optional` has no attribute `__getitem__`.
            return "@" in self.filename_spec
        else:
            return False



# Stubs for email.mime.base (Python 3.4)

import email.message
import sys
from email.policy import Policy
from typing import Optional, Tuple, Union

_ParamsType = Union[str, None, Tuple[str, Optional[str], str]]

class MIMEBase(email.message.Message):
    if sys.version_info >= (3, 6):
        def __init__(self, _maintype: str, _subtype: str, *, policy: Optional[Policy] = ..., **_params: _ParamsType) -> None: ...
    else:
        def __init__(self, _maintype: str, _subtype: str, **_params: _ParamsType) -> None: ...



# pyre-unsafe

from typing import Any, Optional

from django.http.response import HttpResponseRedirect

def redirect_to_login(
    next: Any, login_url: Optional[Any] = ..., redirect_field_name: str = ...
) -> HttpResponseRedirect: ...



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import os
from functools import wraps
from typing import Optional

import click
import click_log
import IPython
from click import Parameter, Path, argument, option
from traitlets.config import Config

from .analysis_output import AnalysisOutput
from .application import start_app
from .context import Context, pass_context
from .create_database import CreateDatabase
from .database_saver import DatabaseSaver
from .db import DB
from .extensions import prompt_extension
from .filesystem import find_root
from .interactive import Interactive
from .model_generator import ModelGenerator
from .models import PrimaryKeyGenerator
from .pipeline import Pipeline
from .trim_trace_graph import TrimTraceGraph


MARKER_DIRECTORIES = [".pyre", ".hg", ".git", ".svn"]

logger = logging.getLogger("sapp")
logging.basicConfig(format="%(asctime)s [%(levelname)s] %(message)s")


def require_option(current_ctx: click.Context, param_name: str) -> None:
    """Throw an exception if an option wasn't required. This is useful when its
    optional in some contexts but required for a subcommand"""

    ctx = current_ctx
    param_definition = None
    while ctx is not None:
        # ctx.command.params has the actual definition of the param. We use
        # this when raising the exception.
        param_definition = next(
            (p for p in ctx.command.params if p.name == param_name), None
        )

        # ctx.params has the current value of the parameter, as set by the user.
        if ctx.params.get(param_name):
            return
        ctx = ctx.parent

    assert param_definition, f"unknown parameter {param_name}"
    raise click.MissingParameter(ctx=current_ctx, param=param_definition)


def common_options(func):
    @click.group(context_settings={"help_option_names": ["--help", "-h"]})
    @click_log.simple_verbosity_option(logger)
    @option(
        "--repository",
        "-r",
        default=lambda: find_root(MARKER_DIRECTORIES),
        type=Path(exists=True, file_okay=False),
        help="Root of the repository (regardless of the directory analyzed)",
    )
    @option(
        "--database-name",
        "--dbname",
        callback=default_database,
        type=Path(dir_okay=False),
    )
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def default_database(ctx: click.Context, _param: Parameter, value: Optional[str]):
    """Try to guess a reasonable database name by looking at the repository path"""
    if value:
        return value

    if ctx.params["repository"]:
        return os.path.join(ctx.params["repository"], DB.DEFAULT_DB_FILE)

    raise click.BadParameter("Could not guess a database location")


@click.command(
    help="interactive exploration of issues",
    context_settings={"ignore_unknown_options": True},
)
@pass_context
@click.argument("ipython_args", nargs=-1, type=click.UNPROCESSED)
def explore(ctx: Context, ipython_args):
    scope_vars = Interactive(
        database=ctx.database,
        # pyre-fixme[6]: Expected `str` for 2nd param but got `Optional[str]`.
        repository_directory=ctx.repository,
        parser_class=ctx.parser_class,
    ).setup()
    config = Config()
    config.InteractiveShellApp.extensions = [
        prompt_extension.__name__
    ] + ctx.ipython_extensions
    config.InteractiveShellApp.profile = "sapp"
    config.InteractiveShellApp.display_banner = False

    config.InteractiveShell.show_rewritten_input = False
    config.InteractiveShell.autocall = 2

    IPython.start_ipython(
        argv=ipython_args if ipython_args else [], user_ns=scope_vars, config=config
    )


@click.command(help="parse static analysis output and save to disk")
@pass_context
@option("--run-kind", type=str)
@option("--branch", type=str)
@option("--commit-hash", type=str)
@option("--job-id", type=str)
@option("--differential-id", type=int)
@option(
    "--previous-issue-handles",
    type=Path(exists=True),
    help=(
        "file containing list of issue handles to compare INPUT_FILE to "
        "(preferred over --previous-input)"
    ),
)
@option(
    "--previous-input",
    type=Path(exists=True),
    help="static analysis output to compare INPUT_FILE to",
)
@option(
    "--linemap",
    type=Path(exists=True),
    help="json file mapping new locations to old locations",
)
@option(
    "--store-unused-models",
    is_flag=True,
    help="store pre/post conditions unrelated to an issue",
)
@argument("input_file", type=Path(exists=True))
def analyze(
    ctx: Context,
    run_kind,
    branch,
    commit_hash,
    job_id,
    differential_id,
    previous_issue_handles,
    previous_input,
    linemap,
    store_unused_models,
    input_file,
):
    # Store all options in the right places
    summary_blob = {
        "run_kind": run_kind,
        "repository": ctx.repository,
        "branch": branch,
        "commit_hash": commit_hash,
        "old_linemap_file": linemap,
        "store_unused_models": store_unused_models,
    }

    if job_id is None and differential_id is not None:
        job_id = "user_input_" + str(differential_id)
    summary_blob["job_id"] = job_id

    if previous_issue_handles:
        summary_blob["previous_issue_handles"] = AnalysisOutput.from_file(
            previous_issue_handles
        )
    elif previous_input:
        previous_input = AnalysisOutput.from_file(previous_input)

    # Construct pipeline
    input_files = (AnalysisOutput.from_file(input_file), previous_input)
    pipeline_steps = [
        ctx.parser_class(),
        CreateDatabase(ctx.database),
        ModelGenerator(),
        TrimTraceGraph(),
        # pyre-fixme[6]: Expected `bool` for 2nd param but got `PrimaryKeyGenerator`.
        DatabaseSaver(ctx.database, PrimaryKeyGenerator()),
    ]
    # pyre-fixme[6]: Expected
    #  `List[tools.sapp.sapp.pipeline.PipelineStep[typing.Any, typing.Any]]` for 1st
    #  param but got `List[typing.Union[DatabaseSaver, ModelGenerator, TrimTraceGraph,
    #  tools.sapp.sapp.base_parser.BaseParser]]`.
    pipeline = Pipeline(pipeline_steps)
    pipeline.run(input_files, summary_blob)


@click.command(
    help="backend flask server for exploration of issues",
    context_settings={"ignore_unknown_options": True},
)
@pass_context
def server(ctx: Context):
    start_app(ctx.database)


commands = [analyze, explore, server]



# pyre-unsafe

from typing import Any, ContextManager

class override_settings(ContextManager[None]):
    def __call__(self, test_func: Any) -> Any: ...
    def __enter__(self) -> None: ...
    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None: ...



import sys
from typing import Generic, Iterable, Optional, Tuple, TypeVar

_S = TypeVar("_S")
_SLT = TypeVar("_SLT", int, float, bool, str, bytes, None)

if sys.version_info >= (3, 8):
    class SharedMemory:
        def __init__(self, name: Optional[str] = ..., create: bool = ..., size: int = ...) -> None: ...
        @property
        def buf(self) -> memoryview: ...
        @property
        def name(self) -> str: ...
        @property
        def size(self) -> int: ...
        def close(self) -> None: ...
        def unlink(self) -> None: ...
    class ShareableList(Generic[_SLT]):
        shm: SharedMemory
        def __init__(self, sequence: Optional[Iterable[_SLT]] = ..., *, name: Optional[str] = ...) -> None: ...
        def __getitem__(self, position: int) -> _SLT: ...
        def __setitem__(self, position: int, value: _SLT) -> None: ...
        def __reduce__(self: _S) -> Tuple[_S, Tuple[_SLT, ...]]: ...
        def __len__(self) -> int: ...
        @property
        def format(self) -> str: ...
        def count(self, value: _SLT) -> int: ...
        def index(self, value: _SLT) -> int: ...
